#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Constants Module
"""

from math import sqrt
from datetime import date

c_vac = 299792.458
k_0 = 0.9996


# Ellipsoid constants
class Ellipsoid(object):
    def __init__(self, semimaj, inversef):
        self.semimaj = semimaj
        self.inversef = inversef
        self.f = 1 / self.inversef
        self.semimin = float(self.semimaj * (1 - self.f))
        self.ecc1sq = float(self.f * (2 - self.f))
        self.ecc2sq = float(self.ecc1sq / (1 - self.ecc1sq))
        self.ecc1 = sqrt(self.ecc1sq)
        self.n = float(self.f / (2 - self.f))
        self.n2 = self.n ** 2
        self.meanradius = (2 * self.semimaj + self.semimin)/3


# Geodetic Reference System 1980
# www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7019
grs80 = Ellipsoid(6378137, 298.257222101)


# World Geodetic System 1984
# www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7030
wgs84 = Ellipsoid(6378137, 298.257223563)


# Australian National Spheroid
# www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7003
ans = Ellipsoid(6378160, 298.25)


# International (Hayford) 1924
# www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7022
intl24 = Ellipsoid(6378388, 297)


# Projections
class Projection(object):
    def __init__(self, falseeast, falsenorth, cmscale, zonewidth, initialcm):
        """
        Transverse Mercator Projection Parameters
        :param falseeast: Easting (m) assigned to Central Meridian
        :param falsenorth: Northing (m) assigned to Equator
        :param cmscale: Central Meridian Scale Factor (unitless, 1 is no scale)
        :param zonewidth: Width (decimal degrees) of each TM Zone
        :param initialcm: Longitude (decimal degrees) of TM Zone 1
        """
        self.falseeast = falseeast
        self.falsenorth = falsenorth
        self.cmscale = cmscale
        self.zonewidth = zonewidth
        self.initialcm = initialcm


utm = Projection(500000, 10000000, 0.9996, 6, -177)

# Integrated Survey Grid - used in NSW as the projection for AGD66
# Spatial Services projections page - https://www.spatial.nsw.gov.au/surveying/geodesy/projections
# ISG Technical Manual - https://www.spatial.nsw.gov.au/__data/assets/pdf_file/0017/25730/ISG.pdf
isg = Projection(300000, 5000000, 0.99994, 2, -177)


# Helmert 14 parameter transformation
class Transformation(object):
    def __init__(self, from_datum, to_datum, ref_epoch,
                 tx, ty, tz, sc, rx, ry, rz,
                 d_tx=0.0, d_ty=0.0, d_tz=0.0, d_sc=0.0,
                 d_rx=0.0, d_ry=0.0, d_rz=0.0, tf_sd=None):
        self.from_datum = from_datum   # Datum transforming from
        self.to_datum = to_datum       # Datum transforming to
        self.ref_epoch = ref_epoch     # Ref. epoch (datetime.date object)
        self.tx = tx                   # Translation in X (m)
        self.ty = ty                   # Translation in Y (m)
        self.tz = tz                   # Translation in Z (m)
        self.sc = sc                   # Scale factor (ppm)
        self.rx = rx                   # Rotation about X (arcsec)
        self.ry = ry                   # Rotation about Y (arcsec)
        self.rz = rz                   # Rotation about Z (arcsec)
        self.d_tx = d_tx               # Translation rate of change in X (m/yr)
        self.d_ty = d_ty               # Translation rate of change in Y (m/yr)
        self.d_tz = d_tz               # Translation rate of change in Z (m/yr)
        self.d_sc = d_sc               # Scale factor rate of change (ppm/yr)
        self.d_rx = d_rx               # Rot rate of change about X (arcsec/yr)
        self.d_ry = d_ry               # Rot rate of change about Y (arcsec/yr)
        self.d_rz = d_rz               # Rot rate of change about Z (arcsec/yr)
        self.tf_sd = tf_sd             # TransformationSD object

    def __repr__(self):
        return ('Transformation: '
                + 'From ' + repr(self.from_datum) +
                ' to ' + repr(self.to_datum) + '\n'
                + 'Reference Epoch: ' + repr(self.ref_epoch) + '\n'
                + '  tx: ' + repr(self.tx) + 'm + ' + repr(self.d_tx) +
                'm/yr' + '\n'
                + '  ty: ' + repr(self.ty) + 'm + ' + repr(self.d_ty) +
                'm/yr' + '\n'
                + '  tz: ' + repr(self.tz) + 'm + ' + repr(self.d_tz) +
                'm/yr' + '\n'
                + '  sc: ' + repr(self.sc) + 'ppm + ' + repr(self.d_sc) +
                'ppm/yr' + '\n'
                + '  rx: ' + repr(self.rx) + '\" + ' + repr(self.d_rx) +
                '\"/yr' + '\n'
                + '  ry: ' + repr(self.ry) + '\" + ' + repr(self.d_ry) +
                '\"/yr' + '\n'
                + '  rz: ' + repr(self.rz) + '\" + ' + repr(self.d_rz) +
                '\"/yr' + '\n')

    def __neg__(self):
        """
        Reverses the direction of the transformation object, i.e.,
        the ITRF2014 to ITRF2000 transformation becomes the ITRF2000 to
        ITRF2014 transformation.

        :return: reversed direction transformation object
        """
        return Transformation(self.to_datum,
                              self.from_datum,
                              self.ref_epoch,
                              -self.tx, -self.ty, -self.tz,
                              -self.sc,
                              -self.rx, -self.ry, -self.rz,
                              -self.d_tx, -self.d_ty, -self.d_tz,
                              -self.d_sc,
                              -self.d_rx, -self.d_ry, -self.d_rz,
                              tf_sd=self.tf_sd
                              )

    def __add__(self, other):
        """
        Change the transformation epoch to a specified date. Advances all
        transformation parameters by their respective rates of change.

        :param other: datetime.date object
        :return: transformation object with parameters and ref epoch moved to
        the specified date
        """
        if type(other) == date:
            timediff = (other - self.ref_epoch).days/365.25

            if type(self.tf_sd) == TransformationSD:
                self.tf_sd.sd_tx = (self.tf_sd.sd_tx**2 + (self.tf_sd.sd_d_tx * timediff)**2) ** 0.5
                self.tf_sd.sd_ty = (self.tf_sd.sd_ty**2 + (self.tf_sd.sd_d_ty * timediff)**2) ** 0.5
                self.tf_sd.sd_tz = (self.tf_sd.sd_tz**2 + (self.tf_sd.sd_d_tz * timediff)**2) ** 0.5
                self.tf_sd.sd_sc = (self.tf_sd.sd_sc**2 + (self.tf_sd.sd_d_sc * timediff)**2) ** 0.5
                self.tf_sd.sd_rx = (self.tf_sd.sd_rx**2 + (self.tf_sd.sd_d_rx * timediff)**2) ** 0.5
                self.tf_sd.sd_ry = (self.tf_sd.sd_ry**2 + (self.tf_sd.sd_d_ry * timediff)**2) ** 0.5
                self.tf_sd.sd_rz = (self.tf_sd.sd_rz**2 + (self.tf_sd.sd_d_rz * timediff)**2) ** 0.5

            return Transformation(self.to_datum,
                                  self.from_datum,
                                  other,
                                  round(self.tx + (self.d_tx * timediff), 8),
                                  round(self.ty + (self.d_ty * timediff), 8),
                                  round(self.tz + (self.d_tz * timediff), 8),
                                  round(self.sc + (self.d_sc * timediff), 8),
                                  round(self.rx + (self.d_rx * timediff), 8),
                                  round(self.ry + (self.d_ry * timediff), 8),
                                  round(self.rz + (self.d_rz * timediff), 8),
                                  self.d_tx,
                                  self.d_ty,
                                  self.d_tz,
                                  self.d_sc,
                                  self.d_rx,
                                  self.d_ry,
                                  self.d_rz,
                                  self.tf_sd
                                  )
        else:
            ValueError('supports adding datetime.date objects only')


class TransformationSD(object):
    def __init__(self, sd_tx=None, sd_ty=None, sd_tz=None, sd_sc=None,
                 sd_rx=None, sd_ry=None, sd_rz=None,
                 sd_d_tx=None, sd_d_ty=None, sd_d_tz=None, sd_d_sc=None,
                 sd_d_rx=None, sd_d_ry=None, sd_d_rz=None):
        self.sd_tx = sd_tx         # one-sigma uncertainty of tx (m)
        self.sd_ty = sd_ty         # one-sigma uncertainty of ty (m)
        self.sd_tz = sd_tz         # one-sigma uncertainty of tz (m)
        self.sd_sc = sd_sc         # one-sigma uncertainty of sc (ppm)
        self.sd_rx = sd_rx         # one-sigma uncertainty of rx (arcsec/yr)
        self.sd_ry = sd_ry         # one-sigma uncertainty of ry (arcsec/yr)
        self.sd_rz = sd_rz         # one-sigma uncertainty of rz (arcsec/yr)
        self.sd_d_tx = sd_d_tx     # one-sigma uncertainty of d_tx (m/yr)
        self.sd_d_ty = sd_d_ty     # one-sigma uncertainty of d_ty (m/yr)
        self.sd_d_tz = sd_d_tz     # one-sigma uncertainty of d_tz (m/yr)
        self.sd_d_sc = sd_d_sc     # one-sigma uncertainty of d_sc (ppm/yr)
        self.sd_d_rx = sd_d_rx     # one-sigma uncertainty of d_rx (arcsec/yr)
        self.sd_d_ry = sd_d_ry     # one-sigma uncertainty of d_ry (arcsec/yr)
        self.sd_d_rz = sd_d_rz     # one-sigma uncertainty of d_rz (arcsec/yr)


def iers2trans(itrf_from, itrf_to, ref_epoch, tx, ty, tz, sc, rx, ry, rz,
               d_tx, d_ty, d_tz, d_sc, d_rx, d_ry, d_rz):
    """
    Used to convert IERS transformation parameters into GeodePy Transformation
    class parameters.
    Note: All rotation and delta rotation terms have the sign change applied.
    :param itrf_from: ITRF realization transforming from
    :param itrf_to: ITRF realization transforming to
    :param ref_epoch: Reference epoch (YYYY.DOY)
    :param tx: Translation in X (mm)
    :param ty: Translation in Y (mm)
    :param tz: Translation in Z (mm)
    :param sc: Scale factor (ppb)
    :param rx: Rotation about X (milliarcsec)
    :param ry: Rotation about Y (milliarcsec)
    :param rz: Rotation about Z (milliarcsec)
    :param d_tx: Translation rate of change in X (mm/yr)
    :param d_ty: Translation rate of change in Y (mm/yr)
    :param d_tz: Translation rate of change in Z (mm/yr)
    :param d_sc: Scale factor rate of change (ppb/yr)
    :param d_rx: Rate of change in Rotation about X (milliarcsec/yr)
    :param d_ry: Rate of change in Rotation about X (milliarcsec/yr)
    :param d_rz: Rate of change in Rotation about X (milliarcsec/yr)
    :return: Transformation object following the Australian convention
    """
    return Transformation(itrf_from, itrf_to, ref_epoch,
                          round(tx / 1000, 8), round(ty / 1000, 8),
                          round(tz / 1000, 8), round(sc / 1000, 8),
                          round(-rx / 1000, 8), round(-ry / 1000, 8),
                          round(-rz / 1000, 8), round(d_tx / 1000, 8),
                          round(d_ty / 1000, 8), round(d_tz / 1000, 8),
                          round(d_sc / 1000, 8), round(-d_rx / 1000, 8),
                          round(-d_ry / 1000, 8), round(-d_rz / 1000, 8))


"""
Transformation Parameter Definitions

Please note transformation parameter definitions should align to the following
GeodePy conventions:

Transformation Variable Naming:
datumYYYY_to_datumYYYY[_suffix]
where:
datum - common abbreviation of datum name
YYYY - year (last two digits of year if before 2000, 4 digits if 2000 or later)
Suffix - optional extension used to differentiate between regional
transformation parameters

Transformation Standard Deviation Variable Naming:
datumYYYY_to_datumYYYY[_suffix]_sd

All parameters in a transformation variable definition should be explicitly
named rather than positionally defined only, as this eases manual checking
of parameters against sources. For example, the from datum should be defined
as from_datum='datumYYYY' rather than just 'datumYYYY'.

Reverse transformations should be explicitly defined using the negative
overload, i.e:
datum2YYYY_to_datum1YYYY = -datum1YYYY_to_datum2YYYY

References to the source of the transformation(s) parameters should be given,
including a web URL
"""
# GDA94 to GDA2020 transformation parameters [GDA2020 Tech Manual v1.2]
# Ref: https://www.icsm.gov.au/gda2020-and-gda94-technical-manuals

gda94_to_gda2020_sd = TransformationSD(
    sd_tx=0.0007, sd_ty=0.0006, sd_tz=0.0007,
    sd_sc=0.00010,
    sd_rx=0.000011, sd_ry=0.000010, sd_rz=0.000011,)

gda94_to_gda2020 = Transformation(
    from_datum='GDA94', to_datum='GDA2020', ref_epoch=0,
    tx=0.06155, ty=-0.01087, tz=-0.04019,
    sc=-0.009994,
    rx=-0.0394924, ry=-0.0327221, rz=-0.0328979,
    tf_sd=gda94_to_gda2020_sd)

gda2020_to_gda94 = -gda94_to_gda2020


# ITRF2014 to GDA2020 transformation parameters [GDA2020 Tech Manual v1.2].
# This transformation is also called the Australian Plate Motion Model and it
# was derived using the 109 ARGN and AuScope GNSS CORS that were used to define
# the National Measurement (Recognized-Value Standard of Measurement of
# Position)Determination 2017.
# Ref: https://www.legislation.gov.au/Details/F2017L01352

itrf2014_to_gda2020_sd = TransformationSD(
    sd_tx=0.00, sd_ty=0.00, sd_tz=0.00,
    sd_sc=0.00,
    sd_rx=0.00, sd_ry=0.00, sd_rz=0.00,
    sd_d_tx=0.00, sd_d_ty=0.00, sd_d_tz=0.00,
    sd_d_sc=0.00,
    sd_d_rx=0.00000417, sd_d_ry=0.00000401, sd_d_rz=0.00000370)

itrf2014_to_gda2020 = Transformation(
    from_datum='ITRF2014', to_datum='GDA2020', ref_epoch=date(2020, 1, 1),
    tx=0, ty=0, tz=0, sc=0, rx=0, ry=0, rz=0,
    d_tx=0, d_ty=0, d_tz=0, d_sc=0,
    d_rx=0.00150379, d_ry=0.00118346, d_rz=0.00120716,
    tf_sd=itrf2014_to_gda2020_sd)

gda2020_to_itrf2014 = -itrf2014_to_gda2020


# ATRF2014 to GDA2020 transformation parameters. Note this transformation is
# identical to the above Australian Plate Motion Model
# Ref: https://www.icsm.gov.au/australian-terrestrial-reference-frame

atrf2014_to_gda2020_sd = TransformationSD(
    sd_tx=0.00, sd_ty=0.00, sd_tz=0.00,
    sd_sc=0.00,
    sd_rx=0.00, sd_ry=0.00, sd_rz=0.00,
    sd_d_tx=0.00, sd_d_ty=0.00, sd_d_tz=0.00,
    sd_d_sc=0.00,
    sd_d_rx=0.00000417, sd_d_ry=0.00000401, sd_d_rz=0.00000370)

atrf2014_to_gda2020 = Transformation(
    from_datum='ATRF2014', to_datum='GDA2020', ref_epoch=date(2020, 1, 1),
    tx=0, ty=0, tz=0, sc=0, rx=0, ry=0, rz=0,
    d_tx=0, d_ty=0, d_tz=0, d_sc=0,
    d_rx=0.00150379, d_ry=0.00118346, d_rz=0.00120716,
    tf_sd=atrf2014_to_gda2020_sd)


# GDA94 to ITRF transformation parameters [Dawson and Woods (2010)]
# AGD66 and AGD84 to GDA94 transformation parameters [GDA94 Tech Manual v2.4]
# http://www.icsm.gov.au/datum/gda2020-and-gda94-technical-manuals

# Transformation Standard Deviations
itrf2008_to_gda94_sd = TransformationSD(
    sd_tx=0.00091, sd_ty=0.00078, sd_tz=0.00106,
    sd_sc=0.000126,
    sd_rx=0.0000221, sd_ry=0.0000236, sd_rz=0.0000194,
    sd_d_tx=0.00008, sd_d_ty=0.00007, sd_d_tz=0.00011,
    sd_d_sc=0.000013,
    sd_d_rx=0.0000028, sd_d_ry=0.0000030, sd_d_rz=0.0000023)

itrf2005_to_gda94_sd = TransformationSD(
    sd_tx=0.00256, sd_ty=0.00187, sd_tz=0.00337,
    sd_sc=0.000227,
    sd_rx=0.0000883, sd_ry=0.0000972, sd_rz=0.0000600,
    sd_d_tx=0.00028, sd_d_ty=0.00020, sd_d_tz=0.00036,
    sd_d_sc=0.000022,
    sd_d_rx=0.0000096, sd_d_ry=0.0000106, sd_d_rz=0.0000070)

itrf2000_to_gda94_sd = TransformationSD(
    sd_tx=0.00874, sd_ty=0.00412, sd_tz=0.01105,
    sd_sc=0.000423,
    sd_rx=0.0002852, sd_ry=0.0003602, sd_rz=0.0001567,
    sd_d_tx=0.00166, sd_d_ty=0.00083, sd_d_tz=0.00209,
    sd_d_sc=0.000076,
    sd_d_rx=0.0000537, sd_d_ry=0.0000677, sd_d_rz=0.0000317)

itrf97_to_gda94_sd = TransformationSD(
    sd_tx=0.01107, sd_ty=0.00574, sd_tz=0.01427,
    sd_sc=0.000425,
    sd_rx=0.0003757, sd_ry=0.0004642, sd_rz=0.0001955,
    sd_d_tx=0.00276, sd_d_ty=0.00136, sd_d_tz=0.00371,
    sd_d_sc=0.000101,
    sd_d_rx=0.0000967, sd_d_ry=0.0001196, sd_d_rz=0.0000453)

itrf96_to_gda94_sd = TransformationSD(
    sd_tx=0.02033, sd_ty=0.01059, sd_tz=0.02866,
    sd_sc=0.000653,
    sd_rx=0.0007627, sd_ry=0.0009049, sd_rz=0.0003225,
    sd_d_tx=0.00753, sd_d_ty=0.00391, sd_d_tz=0.01074,
    sd_d_sc=0.000228,
    sd_d_rx=0.0002850, sd_d_ry=0.0003382, sd_d_rz=0.0001169)

# Transformations
itrf2008_to_gda94 = Transformation(
    from_datum='ITRF2008', to_datum='GDA94', ref_epoch=date(1994, 1, 1),
    tx=-0.08468, ty=-0.01942, tz=0.03201,
    sc=0.00971,
    rx=-0.0004254, ry=0.0022578, rz=0.0024015,
    d_tx=0.00142, d_ty=0.00134, d_tz=0.00090,
    d_sc=0.000109,
    d_rx=0.0015461, d_ry=0.0011820, d_rz=0.0011551,
    tf_sd=itrf2008_to_gda94_sd)

itrf2005_to_gda94 = Transformation(
    from_datum='ITRF2005', to_datum='GDA94', ref_epoch=date(1994, 1, 1),
    tx=-0.07973, ty=-0.00686, tz=0.03803,
    sc=0.006636,
    rx=-0.0000351, ry=0.0021211, rz=0.0021411,
    d_tx=0.00225, d_ty=-0.00062, d_tz=-0.00056,
    d_sc=0.000294,
    d_rx=0.0014707, d_ry=0.0011443, d_rz=0.0011701,
    tf_sd=itrf2005_to_gda94_sd)

itrf2000_to_gda94 = Transformation(
    from_datum='ITRF2000', to_datum='GDA94', ref_epoch=date(1994, 1, 1),
    tx=-0.04591, ty=-0.02985, tz=-0.02037,
    sc=0.00707,
    rx=-0.0016705, ry=0.0004594, rz=0.0019356,
    d_tx=-0.00466, d_ty=0.00355, d_tz=0.01124,
    d_sc=0.000249,
    d_rx=0.0017454, d_ry=0.0014868, d_rz=0.001224,
    tf_sd=itrf2000_to_gda94_sd)

itrf97_to_gda94 = Transformation(
    from_datum='ITRF97', to_datum='GDA94', ref_epoch=date(1994, 1, 1),
    tx=-0.01463, ty=-0.02762, tz=-0.02532,
    sc=0.006695,
    rx=-0.0017893, ry=-0.0006047, rz=0.0009962,
    d_tx=-0.00860, d_ty=0.00036, d_tz=0.01125,
    d_sc=0.000007,
    d_rx=0.0016394, d_ry=0.0015198, d_rz=0.0013801,
    tf_sd=itrf97_to_gda94_sd)

itrf96_to_gda94 = Transformation(
    from_datum='ITRF96', to_datum='GDA94', ref_epoch=date(1994, 1, 1),
    tx=0.02454, ty=-0.03643, tz=-0.06812,
    sc=0.006901,
    rx=-0.0027359, ry=-0.0020431, rz=0.0003731,
    d_tx=-0.02180, d_ty=0.00471, d_tz=0.02627,
    d_sc=0.000388,
    d_rx=0.0020203, d_ry=0.0021735, d_rz=0.0016290,
    tf_sd=itrf96_to_gda94_sd)

agd84_to_gda94 = Transformation(
    from_datum='AGD84', to_datum='GDA94', ref_epoch=0,
    tx=-117.763, ty=-51.510, tz=139.061,
    sc=-0.191,
    rx=-0.292, ry=-0.443, rz=-0.277)

agd66_to_gda94 = Transformation(
    from_datum='AGD66', to_datum='GDA94', ref_epoch=0,
    tx=-117.808, ty=-51.536, tz=137.784,
    sc=-0.290,
    rx=-0.303, ry=-0.446, rz=-0.234)

agd66_to_gda94_act = Transformation(
    from_datum='AGD66', to_datum='GDA94', ref_epoch=0,
    tx=-129.193, ty=-41.212, tz=130.730,
    sc=-2.955,
    rx=-0.246, ry=-0.374, rz=-0.329)

agd66_to_gda94_tas = Transformation(
    from_datum='AGD66', to_datum='GDA94', ref_epoch=0,
    tx=-120.271, ty=-64.543, tz=161.632,
    sc=2.499,
    rx=-0.217, ry=0.067, rz=0.129)

agd66_to_gda94_vicnsw = Transformation(
    from_datum='AGD66', to_datum='GDA94', ref_epoch=0,
    tx=-119.353, ty=-48.301, tz=139.484,
    sc=-0.613,
    rx=-0.415, ry=-0.260, rz=-0.437)

agd66_to_gda94_nt = Transformation(
    from_datum='AGD66', to_datum='GDA94', ref_epoch=0,
    tx=-124.133, ty=-42.003, tz=137.400,
    sc=-1.854,
    rx=0.008, ry=-0.557, rz=-0.178)

gda94_to_itrf2008 = -itrf2008_to_gda94
gda94_to_itrf2005 = -itrf2005_to_gda94
gda94_to_itrf2000 = -itrf2000_to_gda94
gda94_to_itrf97 = -itrf97_to_gda94
gda94_to_itrf96 = -itrf96_to_gda94
gda94_to_agd84 = -agd84_to_gda94
gda94_to_agd66 = -agd66_to_gda94
gda94_to_agd66_act = -agd66_to_gda94_act
gda94_to_agd66_tas = -agd66_to_gda94_tas
gda94_to_agd66_vicnsw = -agd66_to_gda94_vicnsw
gda94_to_agd66_nt = -agd66_to_gda94_nt

# ITRF2020 parameters
# Note: These parameter definitions use the units, rotation and delta-rotation
# sign convention used by the IERS. Units are converted and rotations flipped in
# the transformation object to allow compatibility with GeodePy functions.
# Ref: https://itrf.ign.fr/docs/solutions/itrf2020/Transfo-ITRF2020_TRFs.txt

itrf2020_to_itrf2014_vel = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF2014', ref_epoch=date(2015, 1, 1),
    tx=0.0, ty=-0.1, tz=0.2,
    sc=-0.42,
    rx=0.0, ry=0.0, rz=0.0,
    d_tx=0.0, d_ty=0.0, d_tz=0.0,
    d_sc=0.0,
    d_rx=0.0, d_ry=0.0, d_rz=0.0)

itrf2020_to_itrf2014 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF2014', ref_epoch=date(2015, 1, 1),
    tx=-1.4, ty=-0.9, tz=1.4,
    sc=-0.42,
    rx=0.0, ry=0.0, rz=0.0,
    d_tx=0.0, d_ty=-0.1, d_tz=0.2,
    d_sc=0.0,
    d_rx=0.0, d_ry=0.0, d_rz=0.0)

itrf2020_to_itrf2008 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF2008', ref_epoch=date(2015, 1, 1),
    tx=0.2, ty=1.0, tz=3.3,
    sc=-0.29,
    rx=0.0, ry=0.0, rz=0.0,
    d_tx=0.0, d_ty=-0.1, d_tz=0.1,
    d_sc=0.03,
    d_rx=0.0, d_ry=0.0, d_rz=0.0)

itrf2020_to_itrf2005 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF2005', ref_epoch=date(2015, 1, 1),
    tx=2.7, ty=0.1, tz=-1.4,
    sc=0.65,
    rx=0.0, ry=0.0, rz=0.0,
    d_tx=0.3, d_ty=-0.1, d_tz=0.1,
    d_sc=0.03,
    d_rx=0.0, d_ry=0.0, d_rz=0.0)

itrf2020_to_itrf2000 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF2000', ref_epoch=date(2015, 1, 1),
    tx=-0.2, ty=0.8, tz=-34.2,
    sc=2.25,
    rx=0.0, ry=0.0, rz=0.0,
    d_tx=0.1, d_ty=0.0, d_tz=-1.7,
    d_sc=0.11,
    d_rx=0.0, d_ry=0.0, d_rz=0.0)

itrf2020_to_itrf97 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF97', ref_epoch=date(2015, 1, 1),
    tx=6.5, ty=-3.9, tz=-77.9,
    sc=3.98,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf96 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF96', ref_epoch=date(2015, 1, 1),
    tx=6.5, ty=-3.9, tz=-77.9,
    sc=3.98,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf94 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF94', ref_epoch=date(2015, 1, 1),
    tx=6.5, ty=-3.9, tz=-77.9,
    sc=3.98,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf93 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF93', ref_epoch=date(2015, 1, 1),
    tx=-65.8, ty=1.9, tz=-71.3,
    sc=4.47,
    rx=-3.36, ry=-4.33, rz=0.75,
    d_tx=-2.8, d_ty=-0.2, d_tz=-2.3,
    d_sc=0.12,
    d_rx=-0.11, d_ry=-0.19, d_rz=0.07)

itrf2020_to_itrf92 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF92', ref_epoch=date(2015, 1, 1),
    tx=14.5, ty=-1.9, tz=-85.9,
    sc=3.27,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf91 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF91', ref_epoch=date(2015, 1, 1),
    tx=26.5, ty=12.1, tz=-91.9,
    sc=4.67,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf90 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF90', ref_epoch=date(2015, 1, 1),
    tx=24.5, ty=8.1, tz=-107.9,
    sc=4.97,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf89 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF89', ref_epoch=date(2015, 1, 1),
    tx=29.5, ty=32.1, tz=-145.9,
    sc=8.37,
    rx=0.0, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)

itrf2020_to_itrf88 = iers2trans(
    itrf_from='ITRF2020', itrf_to='ITRF88', ref_epoch=date(2015, 1, 1),
    tx=24.5, ty=-3.9, tz=-169.9,
    sc=11.47,
    rx=0.1, ry=0.0, rz=0.36,
    d_tx=0.1, d_ty=-0.6, d_tz=-3.1,
    d_sc=0.12,
    d_rx=0.0, d_ry=0.0, d_rz=0.02)
    
itrf2014_to_itrf2020 = -itrf2020_to_itrf2014
itrf2008_to_itrf2020 = -itrf2020_to_itrf2008
itrf2005_to_itrf2020 = -itrf2020_to_itrf2005
itrf2000_to_itrf2020 = -itrf2020_to_itrf2000
itrf97_to_itrf2020 = -itrf2020_to_itrf97
itrf96_to_itrf2020 = -itrf2020_to_itrf96
itrf94_to_itrf2020 = -itrf2020_to_itrf94
itrf93_to_itrf2020 = -itrf2020_to_itrf93
itrf92_to_itrf2020 = -itrf2020_to_itrf92
itrf91_to_itrf2020 = -itrf2020_to_itrf91
itrf90_to_itrf2020 = -itrf2020_to_itrf90
itrf89_to_itrf2020 = -itrf2020_to_itrf89
itrf88_to_itrf2020 = -itrf2020_to_itrf88
    

# ITRF2014 parameters
# Note: These parameter definitions use the units, rotation and delta-rotation
# sign convention used by the IERS. Units are converted and rotations flipped in
# the transformation object to allow compatibility with GeodePy functions.
# For more information, see the GDA2020 tech manual section 2.2.1 available
# here: https://www.icsm.gov.au/gda2020-and-gda94-technical-manuals
# Ref: http://itrf.ign.fr/doc_ITRF/Transfo-ITRF2014_ITRFs.txt

itrf2014_to_itrf2008 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF2008', ref_epoch=date(2010, 1, 1),
    tx=1.6, ty=1.9, tz=2.4,
    sc=-0.02,
    rx=0, ry=0, rz=0,
    d_tx=0.0, d_ty=0.0, d_tz=-0.1,
    d_sc=0.03,
    d_rx=0, d_ry=0, d_rz=0)

itrf2014_to_itrf2005 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF2005', ref_epoch=date(2010, 1, 1),
    tx=2.6, ty=1.0, tz=-2.3,
    sc=0.92,
    rx=0, ry=0, rz=0,
    d_tx=0.3, d_ty=0.0, d_tz=-0.1,
    d_sc=0.03,
    d_rx=0, d_ry=0, d_rz=0)

itrf2014_to_itrf2000 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF2000', ref_epoch=date(2010, 1, 1),
    tx=0.7, ty=1.2, tz=-26.1,
    sc=2.12,
    rx=0, ry=0, rz=0,
    d_tx=0.1, d_ty=0.1, d_tz=-1.9,
    d_sc=0.11,
    d_rx=0, d_ry=0, d_rz=0)

itrf2014_to_itrf97 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF97', ref_epoch=date(2010, 1, 1),
    tx=7.4, ty=-0.5, tz=-62.8,
    sc=3.80,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf96 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF96', ref_epoch=date(2010, 1, 1),
    tx=7.4, ty=-0.5, tz=-62.8,
    sc=3.80,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf94 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF94', ref_epoch=date(2010, 1, 1),
    tx=7.4, ty=-0.5, tz=-62.8,
    sc=3.80,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf93 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF93', ref_epoch=date(2010, 1, 1),
    tx=-50.4, ty=3.3, tz=-60.2,
    sc=4.29,
    rx=-2.81, ry=-3.38, rz=0.40,
    d_tx=-2.8, d_ty=-0.1, d_tz=-2.5,
    d_sc=0.12,
    d_rx=-0.11, d_ry=-0.19, d_rz=0.07)

itrf2014_to_itrf92 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF92', ref_epoch=date(2010, 1, 1),
    tx=15.4, ty=1.5, tz=-70.8,
    sc=3.09,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf91 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF91', ref_epoch=date(2010, 1, 1),
    tx=27.4, ty=15.5, tz=-76.8,
    sc=4.49,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf90 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF90', ref_epoch=date(2010, 1, 1),
    tx=25.4, ty=11.5, tz=-92.8,
    sc=4.79,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf89 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF89', ref_epoch=date(2010, 1, 1),
    tx=30.4, ty=35.5, tz=-130.8,
    sc=8.19,
    rx=0, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2014_to_itrf88 = iers2trans(
    itrf_from='ITRF2014', itrf_to='ITRF88', ref_epoch=date(2010, 1, 1),
    tx=25.4, ty=-0.5, tz=-154.8,
    sc=11.29,
    rx=0.1, ry=0, rz=0.26,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.3,
    d_sc=0.12,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf2014 = -itrf2014_to_itrf2008
itrf2005_to_itrf2014 = -itrf2014_to_itrf2005
itrf2000_to_itrf2014 = -itrf2014_to_itrf2000
itrf97_to_itrf2014 = -itrf2014_to_itrf97
itrf96_to_itrf2014 = -itrf2014_to_itrf96
itrf94_to_itrf2014 = -itrf2014_to_itrf94
itrf93_to_itrf2014 = -itrf2014_to_itrf93
itrf92_to_itrf2014 = -itrf2014_to_itrf92
itrf91_to_itrf2014 = -itrf2014_to_itrf91
itrf90_to_itrf2014 = -itrf2014_to_itrf90
itrf89_to_itrf2014 = -itrf2014_to_itrf89
itrf88_to_itrf2014 = -itrf2014_to_itrf88


# ITRF2008 parameters
# Note: These parameter definitions use the units, rotation and delta-rotation
# sign convention used by the IERS. Units are converted and rotations flipped in
# the transformation object to allow compatibility with GeodePy functions.
# For more information, see the GDA2020 tech manual section 2.2.1 available
# here: https://www.icsm.gov.au/gda2020-and-gda94-technical-manuals
# Ref: http://itrf.ign.fr/doc_ITRF/Transfo-ITRF2008_ITRFs.txt

itrf2008_to_itrf2005 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF2005', ref_epoch=date(2000, 1, 1),
    tx=-2.0, ty=-0.9, tz=-4.7,
    sc=0.94,
    rx=0, ry=0, rz=0,
    d_tx=0.3, d_ty=0.0, d_tz=0.0,
    d_sc=0.0,
    d_rx=0, d_ry=0, d_rz=0)

itrf2008_to_itrf2000 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF2000', ref_epoch=date(2000, 1, 1),
    tx=-1.9, ty=-1.7, tz=-10.5,
    sc=1.34,
    rx=0, ry=0, rz=0,
    d_tx=0.1, d_ty=0.1, d_tz=-1.8,
    d_sc=0.08,
    d_rx=0, d_ry=0, d_rz=0)

itrf2008_to_itrf97 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF97', ref_epoch=date(2000, 1, 1),
    tx=4.8, ty=2.6, tz=-33.2,
    sc=2.92,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf96 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF96', ref_epoch=date(2000, 1, 1),
    tx=4.8, ty=2.6, tz=-33.2,
    sc=2.92,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf94 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF94', ref_epoch=date(2000, 1, 1),
    tx=4.8, ty=2.6, tz=-33.2,
    sc=2.92,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf93 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF93', ref_epoch=date(2000, 1, 1),
    tx=4-24.0, ty=2.4, tz=-38.6,
    sc=3.41,
    rx=-1.71, ry=-1.48, rz=-0.30,
    d_tx=-2.8, d_ty=-0.1, d_tz=-2.4,
    d_sc=0.09,
    d_rx=-0.11, d_ry=-0.19, d_rz=0.07)

itrf2008_to_itrf92 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF92', ref_epoch=date(2000, 1, 1),
    tx=12.8, ty=4.6, tz=-41.2,
    sc=2.21,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf91 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF91', ref_epoch=date(2000, 1, 1),
    tx=24.8, ty=18.6, tz=-47.2,
    sc=3.61,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf90 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF90', ref_epoch=date(2000, 1, 1),
    tx=22.8, ty=14.6, tz=-63.2,
    sc=3.91,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf89 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF89', ref_epoch=date(2000, 1, 1),
    tx=27.8, ty=38.6, tz=-101.2,
    sc=7.31,
    rx=0, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2008_to_itrf88 = iers2trans(
    itrf_from='ITRF2008', itrf_to='ITRF88', ref_epoch=date(2000, 1, 1),
    tx=22.8, ty=2.6, tz=-125.2,
    sc=10.41,
    rx=0.10, ry=0, rz=0.06,
    d_tx=0.1, d_ty=-0.5, d_tz=-3.2,
    d_sc=0.09,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2005_to_itrf2008 = -itrf2008_to_itrf2005
itrf2000_to_itrf2008 = -itrf2008_to_itrf2000
itrf97_to_itrf2008 = -itrf2008_to_itrf97
itrf96_to_itrf2008 = -itrf2008_to_itrf96
itrf94_to_itrf2008 = -itrf2008_to_itrf94
itrf93_to_itrf2008 = -itrf2008_to_itrf93
itrf92_to_itrf2008 = -itrf2008_to_itrf92
itrf91_to_itrf2008 = -itrf2008_to_itrf91
itrf90_to_itrf2008 = -itrf2008_to_itrf90
itrf89_to_itrf2008 = -itrf2008_to_itrf89
itrf88_to_itrf2008 = -itrf2008_to_itrf88


# ITRF2005 parameters
# Note: These parameter definitions use the units, rotation and delta-rotation
# sign convention used by the IERS. Units are converted and rotations flipped in
# the transformation object to allow compatibility with GeodePy functions.
# For more information, see the GDA2020 tech manual section 2.2.1 available
# here: https://www.icsm.gov.au/gda2020-and-gda94-technical-manuals
# Ref: http://itrf.ensg.ign.fr/ITRF_solutions/2005/tp_05-00.php

itrf2005_to_itrf2000 = iers2trans(
    itrf_from='ITRF2005', itrf_to='ITRF2000', ref_epoch=date(2000, 1, 1),
    tx=0.1, ty=-0.8, tz=-5.8,
    sc=0.40,
    rx=0, ry=0, rz=0,
    d_tx=-0.2, d_ty=0.1, d_tz=-1.8,
    d_sc=0.08,
    d_rx=0, d_ry=0, d_rz=0)

itrf2000_to_itrf2005 = -itrf2005_to_itrf2000

# ITRF2000 parameters
# ftp://ftp.iers.org/products/reference-systems/terrestrial/itrf/ITRF.TP
# NOTE: This ref lists translations in centimetres. All other ITRF
# transformations are shown in millimetres.
# NOTE: All translations and rates of translation shown below have been
# converted to millimetres.
itrf2000_to_itrf97 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF97', ref_epoch=date(1997, 1, 1),
    tx=6.7, ty=6.1, tz=-18.5,
    sc=1.55,
    rx=0, ry=0, rz=0,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf96 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF96', ref_epoch=date(1997, 1, 1),
    tx=6.7, ty=6.1, tz=-18.5,
    sc=1.55,
    rx=0, ry=0, rz=0,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf94 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF94', ref_epoch=date(1997, 1, 1),
    tx=6.7, ty=6.1, tz=-18.5,
    sc=1.55,
    rx=0, ry=0, rz=0,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf93 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF93', ref_epoch=date(1988, 1, 1),
    tx=12.7, ty=6.5, tz=-20.9,
    sc=1.95,
    rx=-0.39, ry=0.80, rz=-1.14,
    d_tx=-2.9, d_ty=-0.2, d_tz=-0.6,
    d_sc=0.01,
    d_rx=-0.11, d_ry=-0.19, d_rz=0.07)

itrf2000_to_itrf92 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF92', ref_epoch=date(1988, 1, 1),
    tx=14.7, ty=13.5, tz=-13.9,
    sc=0.75,
    rx=0, ry=0, rz=-0.18,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf91 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF91', ref_epoch=date(1988, 1, 1),
    tx=26.7, ty=27.5, tz=-19.9,
    sc=2.15,
    rx=0, ry=0, rz=-0.18,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf90 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF90', ref_epoch=date(1988, 1, 1),
    tx=24.7, ty=23.5, tz=-35.9,
    sc=2.45,
    rx=0, ry=0, rz=-0.18,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf89 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF89', ref_epoch=date(1988, 1, 1),
    tx=29.7, ty=47.5, tz=-73.9,
    sc=5.85,
    rx=0, ry=0, rz=-0.18,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf2000_to_itrf88 = iers2trans(
    itrf_from='ITRF2000', itrf_to='ITRF88', ref_epoch=date(1988, 1, 1),
    tx=24.7, ty=11.5, tz=-97.9,
    sc=8.95,
    rx=0.10, ry=0, rz=-0.18,
    d_tx=0.0, d_ty=-0.6, d_tz=-1.4,
    d_sc=0.01,
    d_rx=0, d_ry=0, d_rz=0.02)

itrf97_to_itrf2000 = -itrf2000_to_itrf97
itrf96_to_itrf2000 = -itrf2000_to_itrf96
itrf94_to_itrf2000 = -itrf2000_to_itrf94
itrf93_to_itrf2000 = -itrf2000_to_itrf93
itrf92_to_itrf2000 = -itrf2000_to_itrf92
itrf91_to_itrf2000 = -itrf2000_to_itrf91
itrf90_to_itrf2000 = -itrf2000_to_itrf90
itrf89_to_itrf2000 = -itrf2000_to_itrf89
itrf88_to_itrf2000 = -itrf2000_to_itrf88


# The locations of files used in the height module
aws_server = '/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/'
file_DOV_PV = aws_server + 'AGQG/DOV_PV.tif'
file_DOV_PM = aws_server + 'AGQG/DOV_PM.tif'
file_AG2020 = aws_server + 'AUSGeoid/AUSGeoid2020_RELEASEV20170908.tif'
file_AG2020_STD = aws_server + 'AUSGeoid/AUSGeoid2020_RELEASEV20170908_err.tif'
file_AVWS = aws_server + 'AGQG/AGQG_20201120.tif'
file_AVWS_STD = aws_server + 'AGQG/AGQG_uncertainty_20201120.tif'
file_GRAV_BA = aws_server + 'AGQG/Bouguer_Grav_RELEASE20191107.tif'
file_AG98=aws_server+'AUSGeoid/AUSGeoid98.tif'
file_AG09=aws_server+'AUSGeoid/AUSGeoid09_V1.01.tif'
file_AG98_DOV_PV=aws_server+'AUSGeoid/AUSGeoid98_DOV_PV.tif'
file_AG98_DOV_PM=aws_server+'AUSGeoid/AUSGeoid98_DOV_PM.tif'
file_AG09_DOV_PV=aws_server+'AUSGeoid/AUSGeoid09_DOV_PV_V1.01.tif'
file_AG09_DOV_PM=aws_server+'AUSGeoid/AUSGeoid09_DOV_PM_V1.01.tif'


# GRS80 normal gravity flattening [Moritz, 2000 Section 4]
grs80_ngf = 0.005302440112



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
GNSS Module

In Development

Specific utility functions for SINEX:

    - remove_stns_sinex()
    - remove_velocity_sinex()
    - remove_matrixzeros_sinex()

General functions for reading from SINEX: 

    - list_sinex_blocks()
    - read_sinex_comments() 
    - read_sinex_header_line()
    - read_sinex_custom()
    - read_sinex_estimate()
    - read_sinex_matrix()
    - read_sinex_sites()
    - read_disconts()
    - read_solution_epochs()
    - read_sinex_header_block()
    - read_sinex_file_reference_block()
    - read_sinex_input_acknowledgments_block()
    - read_sinex_solution_statistics_block()
    - read_sinex_site_receiver_block()
    - read_sinex_site_antenna_block()
    - read_sinex_site_gps_phase_center_block()
    - read_sinex_site_eccentricity_block()
    - read_sinex_site_id_block()
    - read_sinex_solution_epochs_block()
    - read_sinex_solution_estimate_block()
    - read_sinex_solution_apriori_block()
    - read_sinex_solution_matrix_estimate_block()
    - read_sinex_solution_matrix_apriori_block()
    - matrix2dataframe_solution_matrix_estimate()
    - sinex2dataframe_solution_estimate()
    - sinex2dataframe_solution_apriori()
    - sinex2dataframe_solution_matrix_estimate()

General functions for writing to SINEX:
    
    - print_sinex_comments()
    - set_creation_time()
    - dataframe2sinex_solution_estimate()
    - dataframe2sinex_solution_apriori()
    - dataframe2matrix_snx_vcv()
    - dataframe2matrix_solution_matrix_estimate()
    - dataframe2sinex_solution_matrix_estimate()
    - writeSINEX()
    
"""

from datetime import datetime
import numpy as np
import pandas as pd
from numpy import zeros, delete
from geodepy.angles import DMSAngle
import re


def list_sinex_blocks(file):
    """This script lists the blocks in a SINEX file

    :param str file: the input SINEX file
    """
    blocks = []
    with open(file) as f:
        for line in f.readlines():
            if line.startswith('+'):
                col = line.split(' ')
                block = col[0].replace('+', '')
                blocks.append(block.strip())
    for block in blocks:
        print(block)

def print_sinex_comments(file):
    """This script prints comments in a SINEX file.

    :param str file: the input SINEX file
    """
    go = False
    with open(file) as f:
        for line in f.readlines():
            if line.startswith('+FILE/COMMENT'):
                go = True
            if go:
                print(line.strip())
            if line.startswith('-FILE/COMMENT'):
                go = False

def read_sinex_comments(file):
    """This function reads comments in a SINEX file.

    :param str file: the input SINEX file
    :return: comments block
    :rtype: list of strings
    """
    comments = []
    go = False
    with open(file) as f:
        for line in f.readlines():
            if line.startswith('+FILE/COMMENT'):
                go = True
            if go:
                comments.append(line.strip())
            if line.startswith('-FILE/COMMENT'):
                go = False

        comments.insert(-1,f"* File created by Geodepy.gnss.py at {datetime.now().strftime('%d-%m-%Y, %H:%M')}")

    return comments

def set_creation_time():
    """This function sets the creation time, in format YY:DDD:SSSSS, for use
    in the SINEX header line

    :return: creation_time
    :rtype: str
    """
    now = datetime.now()
    time_tup = now.timetuple()
    year = str(time_tup.tm_year)[2:]
    doy = time_tup.tm_yday
    doy = '{:03d}'.format(doy)
    seconds = (now - now.replace(hour=0, minute=0, second=0, microsecond=0))\
        .total_seconds()
    seconds = '{:.0f}'.format(seconds)
    creation_time = year + ':' + doy + ':' + seconds

    return creation_time

def read_sinex_header_line(file):
    """This function reads the header line of a SINEX file into a string

    :param str file: the input SINEX file
    :return: header_line
    :rtype: str
    """
    with open(file) as f:
        header_line = f.readline()

    return header_line

def read_sinex_custom(fp, start_line, end_line):
    """Read custom line range from SINEX. Useful for 
    SINEX with irregular formatting or block that has 
    not been accounted for.

    :param str fp: Path to SINEX file.
    :param int start_line:   Beginning line number. 
    :param int end_line:     Finishing line number.
    :return list custom:  List of string(s).
    """

    # Setup
    i = start_line
    j = end_line
    f = fp
    custom = []

    with open(fp) as f:
        for line in f.readlines()[i-1:j]:
             custom.append(line.strip())

    # Return
    return custom

def read_sinex_estimate(file):
    """This function reads in the SOLUTION/ESTIMATE block of a SINEX file. It
    returns estimate, a list of tuples:

    estimate = [(code, soln, refEpoch, staX, staY, staZ, staX_sd, staY_sd,
                 staZ_sd[, velX, velY, velZ, velX_sd, velY_sd, velZ_sd])...]

    where:
        * code is the stations's 4-character ID
        * soln is the segment of the stations's time series
        * refEpoch is the epoch of the solution in the form YY:DOY:SSSSS (YY
            is the two digit year, DOY is day of year, and SSSSS is the time of
            day in seconds
        * sta[XYZ] is the station coordinates in the Cartesian reference frame
        * sta[XYZ]_sd is the standard deviation of the station coordinates in
            the Cartesian reference frame
        * vel[XYZ] is the station velocity in the Cartesian reference frame
        * vel[XYZ]_sd is the standard deviation of the station velocity in the
            Cartesian reference frame

    Velocities are not included in all SINEX files and so are only returned if
    present.

    :param file: the input SINEX file
    :return: estimate
    """

    # Create data structures and set variables
    lines = []
    estimate = []
    velocities = False
    go = False
    code = ''
    soln = ''
    epoch = ''
    stax = ''
    stay = ''
    staz = ''
    stax_sd = ''
    stay_sd = ''
    staz_sd = ''
    velx = ''
    vely = ''
    velz = ''
    velx_sd = ''
    vely_sd = ''
    velz_sd = ''

    # Read the SOLUTION/ESTIMATE block into a list and determine if there is
    # any velocity information
    with open(file) as f:
        for line in f:
            if line[:18] == '-SOLUTION/ESTIMATE':
                break
            if go and line[:11] == '*INDEX TYPE':
                pass
            elif go:
                if line[7:10] == 'VEL':
                    velocities = True
                lines.append(line)
            if line[:18] == '+SOLUTION/ESTIMATE':
                go = True

    for line in lines:
        typ = line[7:11]
        if typ == 'STAX':
            code = line[14:18]
            soln = line[23:26].lstrip()
            epoch = line[27:39]
            stax = float(line[47:68])
            stax_sd = float(line[69:80])
        elif typ == 'STAY':
            stay = float(line[47:68])
            stay_sd = float(line[69:80])
        elif typ == 'STAZ':
            staz = float(line[47:68])
            staz_sd = float(line[69:80])
            if not velocities:
                info = (code, soln, epoch, stax, stay, staz, stax_sd, stay_sd,
                        staz_sd)
                estimate.append(info)
        elif typ == 'VELX':
            velx = float(line[47:68])
            velx_sd = float(line[69:80])
        elif typ == 'VELY':
            vely = float(line[47:68])
            vely_sd = float(line[69:80])
        elif typ == 'VELZ':
            velz = float(line[47:68])
            velz_sd = float(line[69:80])
            info = (code, soln, epoch, stax, stay, staz, stax_sd, stay_sd,
                    staz_sd, velx, vely, velz, velx_sd, vely_sd, velz_sd)
            estimate.append(info)

    return estimate

def read_sinex_matrix(file):
    """This function reads in the SOLUTION/MATRIX_ESTIMATE block of a SINEX
    file. It returns matrix, a list of tuples:

    matrix = [(code, soln, var_x, covar_xy, covar_xz, var_y, covar_yz,
            var_z[, var_v_x, covar_v_xy, covar_v_xz, var_v_y, covar_v_yz,
            var_v_z])...]

    where:
        * code is the stations's 4-character ID
        * soln is the segment of the stations's time series
        * var_x is the variance in the X coordinate
        * covar_xy is the covariance between the X and the Y coordinates
        * covar_xz is the covariance between the X and the Z coordinates
        * var_y is the variance in the Y coordinate
        * covar_yz is the covariance between the Y and the Z coordinates
        * var_z is the variance in the Z coordinate
        * var_v_x is the variance in the X velocity
        * covar_v_xy is the covariance between the X and the Y velocities
        * covar_v_xz is the covariance between the X and the Z velocities
        * var_v_y is the variance in the Y velocity
        * covar_v_yz is the covariance between the Y and the Z velocities
        * var_v_z is the variance in the Z velocity

    Velocities are not included in all SINEX files and so their VCV information
    is only returned if they are present.

    ToDo: 
    1. The above order is only valid if the matrix is upper triangle. If it is
       lower triangle, then the covar_xy is actually the var_y. Will need to fix 
       this when time permits.

    :param file: the input SINEX file
    :return: matrix
    """

    # Read in the codes (station names) and solutions, and check for velocities
    data = read_sinex_estimate(file)
    code = []
    soln = []
    velocities = False
    for station in data:
        code.append(station[0])
        soln.append(station[1])
    if len(data[0]) == 15:
        velocities = True

    # Read the SOLUTION/MATRIX_ESTIMATE block into a list and determine if the
    # matrix is upper or lower triangular
    lines = []
    lower_triangular = False
    go = False
    with open(file) as f:
        for line in f:
            if line[:25] == '-SOLUTION/MATRIX_ESTIMATE':
                break
            if go and line[:12] == '*PARA1 PARA2':
                pass
            elif go:
                lines.append(line)
            if line[:25] == '+SOLUTION/MATRIX_ESTIMATE':
                if line[26] == 'L':
                    lower_triangular = True
                go = True

    # Create an array containing the matrix elements
    if velocities:
        n = 6 * int(len(code))
    else:
        n = 3 * int(len(code))
    element = zeros((n, n))
    matrix = []
    for line in lines:
        col = line.rstrip().split()
        for i in range(2, len(col)):
            element[int(col[0]) - 1][int(col[1]) + i - 3] = float(col[i])
    if velocities:
        if lower_triangular:
            for i in range(len(code)):
                info = (code[i], soln[i], element[6 * i][6 * i],
                        element[6 * i + 1][6 * i],
                        element[6 * i + 1][6 * i + 1],
                        element[6 * i + 2][6 * i],
                        element[6 * i + 2][6 * i + 1],
                        element[6 * i + 2][6 * i + 2],
                        element[6 * i + 3][6 * i + 3],
                        element[6 * i + 4][6 * i + 3],
                        element[6 * i + 4][6 * i + 4],
                        element[6 * i + 5][6 * i + 3],
                        element[6 * i + 5][6 * i + 4],
                        element[6 * i + 5][6 * i + 5])
                matrix.append(info)
        else:
            for i in range(len(code)):
                info = (code[i], soln[i], element[6 * i][6 * i],
                        element[6 * i][6 * i + 1], element[6 * i][6 * i + 2],
                        element[6 * i + 1][6 * i + 1],
                        element[6 * i + 1][6 * i + 2],
                        element[6 * i + 2][6 * i + 2],
                        element[6 * i + 3][6 * i + 3],
                        element[6 * i + 3][6 * i + 4],
                        element[6 * i + 3][6 * i + 5],
                        element[6 * i + 4][6 * i + 4],
                        element[6 * i + 4][6 * i + 5],
                        element[6 * i + 5][6 * i + 5])
                matrix.append(info)
    else:
        if lower_triangular:
            for i in range(len(code)):
                info = (code[i], soln[i], element[3 * i][3 * i],
                        element[3 * i + 1][3 * i],
                        element[3 * i + 1][3 * i + 1],
                        element[3 * i + 2][3 * i],
                        element[3 * i + 2][3 * i + 1],
                        element[3 * i + 2][3 * i + 2])
                matrix.append(info)
        else:
            for i in range(len(code)):
                info = (code[i], soln[i], element[3 * i][3 * i],
                        element[3 * i][3 * i + 1], element[3 * i][3 * i + 2],
                        element[3 * i + 1][3 * i + 1],
                        element[3 * i + 1][3 * i + 2],
                        element[3 * i + 2][3 * i + 2])
                matrix.append(info)

    return matrix

def read_sinex_sites(file):
    """This function reads in the SITE/ID block of a SINEX file. It returns
    sites, a list of tuples:

    sites = [(site, point, domes, obs, station_description, lon, lat, h)]

    where:
        * site is the site code
        * point is the site's point code
        * domes is the site's dome number
        * obs is the observation technique
        * station_description is a free format desciption of the site
        * lon is the approximate longitude of the site as a DMSAngle object
        * lat is the approximate latitude of the site as a DMSAngle object
        * h is the approximate height of the site

    :param file: the input SINEX file
    :return: sites
    """

    # Read the SITE/ID block into a list
    lines = []
    go = False
    with open(file) as f:
        for line in f:
            if line[:8] == '-SITE/ID':
                break
            if go and line[:8] == '*CODE PT':
                pass
            elif go:
                lines.append(line)
            if line[:8] == '+SITE/ID':
                go = True
    sites = []
    for line in lines:
        site = line[1:5]
        point = line[6:8].lstrip()
        domes = line[9:18]
        obs = line[19:20]
        station_description = line[21:43].lstrip()
        lon = DMSAngle(line[44:55].lstrip())
        lat = DMSAngle(line[56:67].lstrip())
        h = float(line[67:73])
        info = (site, point, domes, obs, station_description, lon, lat, h)
        sites.append(info)

    return sites

def read_disconts(file):
    """This function reads in the SOLUTION/DISCONTINUITY block of a
    SINEX file. It returns disconts , a list of tuples:

    sites = [(site, code1, point, code2, start, end, type)]

    where:
        * site is the site code
        * code1 is unknown
        * point is the site's point code
        * code2 is unknown
        * start is the start time for the point code in YY:DOY:SECOD
        * end is the end time for the point code in YY:DOY:SECOD
        * type is the type of discontinuity; P for position or V for
            velocity

    I could not find the format description for this block.

    :param file: the input discontinuities file
    :return: disconts
    """

    # Read the SOLUTION/DISCONTINUITY block into a list
    lines = []
    go = False
    with open(file) as f:
        for line in f:
            if line[:23] == '-SOLUTION/DISCONTINUITY':
                break
            elif go:
                lines.append(line)
            if line[:23] == '+SOLUTION/DISCONTINUITY':
                go = True
    disconts = []
    for line in lines:
        site = line[1:5]
        code1 = line[5:8].lstrip()
        point = line[8:13].lstrip()
        code2 = line[14:15]
        start = line[16:28]
        end = line[29:41]
        p_or_v = line[42:43]
        info = (site, code1, point, code2, start, end, p_or_v)
        disconts.append(info)

    return disconts

def read_solution_epochs(file):
    """This function reads in the SOLUTION/EPOCHS block of a SINEX file.
    It returns epochs, a list of tuples:

    epochs = [(site, point, sol, obs, start, end, mean)]

    where:
        * site is the site code
        * point is the site's point code
        * sol is the solution number at a site/point
        * obs is the observation technique
        * start is the start time for the solution in YY:DOY:SECOD
        * end is the end time for the solution in YY:DOY:SECOD
        * mean is the mean time for the solution in YY:DOY:SECOD

    :param file: the input SINEX file
    :return: epochs
    """

    # Read the SOLUTION/EPOCHS block into a list
    lines = []
    go = False
    with open(file) as f:
        for line in f:
            if line[:16] == '-SOLUTION/EPOCHS':
                break
            if go and line[:8] == '*Code PT':
                pass
            elif go:
                lines.append(line)
            if line[:16] == '+SOLUTION/EPOCHS':
                go = True
    epochs = []
    # Parse each line, create a tuple and add it to the list
    for line in lines:
        site = line[1:5]
        point = line[6:8].lstrip()
        sol = line[9:13].lstrip()
        obs = line[14:15]
        start = line[16:28]
        end = line[29:41]
        mean = line[42:55].rstrip()
        info = (site, point, sol, obs, start, end, mean)
        epochs.append(info)

    return epochs

def read_sinex_header_block(sinex):
    """This function reads in the header block information 
    of a SINEX file (All lines before the SITE/ID block).

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    with open(sinex, 'r') as f:
        next(f)
        line = f.readline()
        while line:
            block.append(line.rstrip())
            line = f.readline()
            if line.startswith('+SITE/ID'):
                break

    return block

def read_sinex_file_reference_block(sinex):
    """This function reads in the +FILE/REFERENCE block 
    of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+FILE/REFERENCE'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-FILE/REFERENCE'):
                break
            line = f.readline()
            
    return block

def read_sinex_input_acknowledgments_block(sinex):
    """This function reads in the +INPUT/ACKNOWLEDGMENTS 
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+INPUT/ACKNOWLEDGMENTS'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-INPUT/ACKNOWLEDGMENTS'):
                break
            line = f.readline()

    return block

def read_sinex_solution_statistics_block(sinex):
    """This function reads in the +SOLUTION/STATISTICS
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/STATISTICS'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/STATISTICS'):
                break
            line = f.readline()

    return block

def read_sinex_site_receiver_block(sinex):
    """This function reads in the +SITE/RECEIVER
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SITE/RECEIVER'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SITE/RECEIVER'):
                break
            line = f.readline()
            
    return block

def read_sinex_site_antenna_block(sinex):
    """This function reads in the +SITE/ANTENNA
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SITE/ANTENNA'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SITE/ANTENNA'):
                break
            line = f.readline()
            
    return block

def read_sinex_site_gps_phase_center_block(sinex):
    """This function reads in the +SITE/GPS_PHASE_CENTER
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SITE/GPS_PHASE_CENTER'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SITE/GPS_PHASE_CENTER'):
                break
            line = f.readline()
            
    return block

def read_sinex_site_eccentricity_block(sinex):
    """This function reads in the +SITE/ECCENTRICITY
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SITE/ECCENTRICITY'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SITE/ECCENTRICITY'):
                break
            line = f.readline()
            
    return block

def read_sinex_site_id_block(sinex):
    """This function reads in the SITE/ID block of a SINEX file
    into list of strings.

    :param str sinex: input SINEX file
    return: block
    """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SITE/ID'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SITE/ID'):
                break
            line = f.readline()

    return block

def read_sinex_solution_epochs_block(sinex):
    """This function reads in the SOLUTION/EPOCHS block of a SINEX file
    into list of strings.

    :param str sinex: input SINEX file
    return: block
    """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/EPOCHS'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/EPOCHS'):
                break
            line = f.readline()
    return block

def read_sinex_solution_estimate_block(sinex):
    """This function reads in the SOLUTION/ESTIMATE block of a SINEX
    file into list of strings.

    :param str sinex: input SINEX file
    return: block
    """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/ESTIMATE'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/ESTIMATE'):
                break
            line = f.readline()
    return block

def read_sinex_solution_apriori_block(sinex):
    """This function reads in the +SOLUTION/APRIORI
    block of a SINEX file.

        :param str sinex: input SINEX file
        return: block
        """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/APRIORI'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/APRIORI'):
                break
            line = f.readline()
            
    return block

def read_sinex_solution_matrix_estimate_block(sinex):
    """This function reads in the SOLUTION/MATRIX_ESTIMATE block of a SINEX
    file into list of strings.

    :param str sinex: input SINEX file
    return: block
    """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/MATRIX_ESTIMATE'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/MATRIX_ESTIMATE'):
                break
            line = f.readline()

    return block

def read_sinex_solution_matrix_apriori_block(sinex):
    """This function reads in the SOLUTION/MATRIX_APRIORI block 
    of a SINEX file into list of strings.

    :param str sinex: input SINEX file
    return: block
    """

    block = []
    go = False
    with open(sinex, 'r') as f:
        line = f.readline()
        while line:
            if line.startswith('+SOLUTION/MATRIX_APRIORI'):
                go = True
            if go:
                block.append(line.rstrip())
            if line.startswith('-SOLUTION/MATRIX_APRIORI'):
                break
            line = f.readline()
            
    return block

def sinex2dataframe_solution_estimate(fp):
    """This function reads in a SINEX file and returns 
    a dataframe of SOLUTION/ESTIMATE block only. 

    :param str sinex: path of input SINEX file
    return: df
    """
    
    # Get lines
    lines = read_sinex_solution_estimate_block(fp)

    # Remove non-data lines
    for l in lines[:]:
        if l.startswith('*'):
            lines.remove(l)
        if l.startswith('+'):
            lines.remove(l)
        if l.startswith('-'):
            lines.remove(l)

    # Split by column
    lines = [i.split() for i in lines]

    # Isolate into vectors
    row = np.int_(list(zip(*lines))[0])
    par = list(zip(*lines))[1]
    code = list(zip(*lines))[2]
    pt = list(zip(*lines))[3]
    soln = list(zip(*lines))[4]
    refEpoch = list(zip(*lines))[5]
    unit = list(zip(*lines))[6]
    s = np.int_(list(zip(*lines))[7])
    est = np.float64(list(zip(*lines))[8])
    sigma = np.float64(list(zip(*lines))[9])

    # Organise into DataFrame
    dict_temp = {
                "row":row,
                "par":par,
                "code":code,
                "pt":pt,
                "soln":soln,
                "refEpoch":refEpoch,
                "unit":unit,
                "s":s,
                "est":est,
                "sigma":sigma,
    }
    df = pd.DataFrame(dict_temp)

    # Return
    return df

def sinex2dataframe_solution_apriori(fp):
    """This function reads in a SINEX file and returns 
    a dataframe of SOLUTION/APRIORI block only. 

    :param str sinex: path of input SINEX file
    return: df
    """
    
    # Get lines
    lines = read_sinex_solution_apriori_block(fp)

    # Remove non-data lines
    for l in lines[:]:
        if l.startswith('*'):
            lines.remove(l)
        if l.startswith('+'):
            lines.remove(l)
        if l.startswith('-'):
            lines.remove(l)

    # Split by column
    lines = [i.split() for i in lines]

    # Isolate into vectors
    row = np.int_(list(zip(*lines))[0])
    par = list(zip(*lines))[1]
    code = list(zip(*lines))[2]
    pt = list(zip(*lines))[3]
    soln = list(zip(*lines))[4]
    refEpoch = list(zip(*lines))[5]
    unit = list(zip(*lines))[6]
    s = np.int_(list(zip(*lines))[7])
    est = np.float64(list(zip(*lines))[8])
    sigma = np.float64(list(zip(*lines))[9])

    # Organise into DataFrame
    dict_temp = {
                "row":row,
                "par":par,
                "code":code,
                "pt":pt,
                "soln":soln,
                "refEpoch":refEpoch,
                "unit":unit,
                "s":s,
                "est":est,
                "sigma":sigma,
    }
    df = pd.DataFrame(dict_temp)

    # Return
    return df

def sinex2dataframe_solution_matrix_estimate(fp):
    """This function reads in a SINEX file and returns 
    a dataframe of SOLUTION/MATRIX_ESTIMATE block only. 

    :param str sinex: path of input SINEX file
    return: df
    """

    # Get lines
    lines = read_sinex_solution_matrix_estimate_block(fp)

    # Remove non-data lines
    for l in lines[:]:
        if l.startswith('*'):
            lines.remove(l)
        if l.startswith('+'):
            lines.remove(l)
        if l.startswith('-'):
            lines.remove(l)

    # Split by column
    lines = [i.split() for i in lines]

    # Pad out lines with NaN
    for i in range(len(lines)):
        if len(lines[i])==5:
            continue
        if len(lines[i]) == 4:
            lines[i].append(np.nan)
        if len(lines[i]) == 3:
            lines[i].append(np.nan)
            lines[i].append(np.nan)
        
    # Isolate into vectors
    row = np.int_(list(zip(*lines))[0])
    col = np.int_(list(zip(*lines))[1])
    q1 = np.float64(list(zip(*lines))[2])
    q2 = np.float64(list(zip(*lines))[3])
    q3 = np.float64(list(zip(*lines))[4])

    # Organise into DataFrame
    dict_temp = {
                "row":row,
                "col":col,
                "q1":q1,
                "q2":q2,
                "q3":q3,
    }
    df = pd.DataFrame(dict_temp)

    # Return
    return df

def dataframe2sinex_solution_estimate(df):
    """This function reads in a dataframe of the 
    SOLUTIONS/ESTIMATE block from a SINEX, then
    converts each row to a string in a list 
    ready for writing to SINEX.

    :param dataframe df: dataframe of SOLUTION/ESTIMATE block
    return: list of strings
    """

    lines =lines = ["+SOLUTION/ESTIMATE"]
    lines.append("*INDEX TYPE__ CODE PT SOLN _REF_EPOCH__ UNIT S __ESTIMATED VALUE____ _STD_DEV___")

    for i in range(len(df.code)):
        l = " %5i %s   %s  %s    %s %s %-3s  %s %21.14e %.5e" % (df.row.iloc[i], df.par.iloc[i], df.code.iloc[i], df.pt.iloc[i], df.soln.iloc[i], df.refEpoch.iloc[i], df.unit.iloc[i], df.s.iloc[i], df.est.iloc[i], df.sigma.iloc[i])
        lines.append(l)
    
    lines.append("-SOLUTION/ESTIMATE")

    # Return
    return lines

def dataframe2sinex_solution_apriori(df):
    """This function reads in a dataframe of the 
    SOLUTION/APRIORI block from a SINEX, then
    converts each row to a string in a list 
    ready for writing to SINEX.

    :param dataframe df: dataframe of SOLUTION/APRIORI block
    return: list of strings
    """

    lines =lines = ["+SOLUTION/APRIORI"]
    lines.append("*INDEX TYPE__ CODE PT SOLN _REF_EPOCH__ UNIT S __ESTIMATED VALUE____ _STD_DEV___")

    for i in range(len(df.code)):
        l = " %5i %s   %s  %s    %s %s %-3s  %s %21.14e %.5e" % (df.row.iloc[i], df.par.iloc[i], df.code.iloc[i], df.pt.iloc[i], df.soln.iloc[i], df.refEpoch.iloc[i], df.unit.iloc[i], df.s.iloc[i], df.est.iloc[i], df.sigma.iloc[i])
        lines.append(l)
    
    lines.append("-SOLUTION/APRIORI")

    # Return
    return lines

def dataframe2matrix_snx_vcv(df, numPar=3):
    """This function converts a dataframe created from a
    SINEX VCV based from read_sinex_matrix(), and converts 
    it to a full VCV without covariances between sites.

    i.e.  xx xy xz          0  0  0
          xy yy yz .  .  .  0  0  0
          xz yz zz          0  0  0
             .     .           .
             .        .        .
             .           .     .
          0  0  0           xx xy xz 
          0  0  0  .  .  .  xy yy yz
          0  0  0           xz yz zz

    :param DataFrame: dataframe formed from read_sinex_matrix():
    return: Numpy matrix of VCV with no covariances between sites. 
    """
    
    # Matrix dimensions
    # - To Do: Handle check for velocities (?)
    par = numPar # (is there a way to automatically determine if velocitie exist?)
    n = len(df.code)*par
    Q = np.zeros((n, n))

    # Matrix elements and rows of dataframe
    i = 0
    j = 0
    r = 0
    while r < len(df.code):

        # variances
        Q[i+0, j+0] = df.xx[r]
        Q[i+1, j+1] = df.yy[r]
        Q[i+2, j+2] = df.zz[r]

        # covariances
        Q[i+1, j+0] = df.xy[r]
        Q[i+0, j+1] = df.xy[r]
        Q[i+2, j+0] = df.xz[r]
        Q[i+0, j+2] = df.xz[r]
        Q[i+2, j+1] = df.yz[r]
        Q[i+1, j+2] = df.yz[r]
        

        # Counter
        r = r+1
        i = i+3
        j = j+3

    # Return
    return Q
    
def dataframe2matrix_solution_matrix_estimate(df, tri="L"):
    """This function reads in a dataframe of the SINEX 
    SOLUTION/MATRIX_ESTIMATE block formed from 
    sinex2dataframe_solution_matrix_estimate(), and then 
    forms the full VCV matrix from that dataframe. 

    :param DataFrame df: dataframe from sinex2dataframe_solution_matrix_estimate().
    :param String tri: String to indicate "upper" or "lower" triagle matrix. 
    return: Numpy matrix of full VCV. 
    """

    # Triangularity of matrix
    triangle = tri

    # Matrix size
    n = df.row.iloc[-1]
    Q = np.zeros((n, n))

    if triangle == "L":

        for i in range(len(df.row)):
            
            # Get matrix indices
            row = int(df.row[i]) - 1
            col = int(df.col[i]) - 1
            
            # Fill PARA2+0
            q1 = df.q1[i]    
            Q[row, col] = q1
            Q[col, row] = q1
            
            # Fill PARA2+1
            q2 = df.q2[i]
            Q[row, col+1] = q2
            Q[col+1, row] = q2
            
            # Fill PARA2+2
            q3 = df.q3[i]
            Q[row, col+2] = q3
            Q[col+2, row] = q3
        
    if triangle == "U":

        for i in range(len(df.row)):

            # Get matrix indices
            row = int(df.row[i]) - 1
            col = int(df.col[i]) - 1

            if df.col[i] < n-1:
            
                # Fill PARA2+0
                q1 = df.q1[i]    
                Q[row, col] = q1
                Q[col, row] = q1
                
                # Fill PARA2+1
                q2 = df.q2[i]
                Q[row, col+1] = q2
                Q[col+1, row] = q2
                
                # Fill PARA2+2
                q3 = df.q3[i]
                Q[row, col+2] = q3
                Q[col+2, row] = q3
            
            if df.col[i] == n-1:
            
                # Fill PARA2+0
                q1 = df.q1[i]    
                Q[row, col] = q1
                Q[col, row] = q1
                
                # Fill PARA2+1
                q2 = df.q2[i]
                Q[row, col+1] = q2
                Q[col+1, row] = q2

            if df.col[i] == n:
            
                # Fill PARA2+0
                q1 = df.q1[i]    
                Q[row, col] = q1
                Q[col, row] = q1

    # Return
    return Q

def matrix2dataframe_solution_matrix_estimate(m, tri="L"):
    """This function reads a VCV in matrix format and writes it
    to dataframe for SINEX SOLUTION/MATRIX_ESTIMATE format. It 
    is the format produced from sinex2dataframe_solution_matrix_estimate().

    :param numpy.array() m: A numpy array of the VCV matrix. 
    :param String tri: String to indicate "upper" or "lower" triagle matrix. 
    return: DataFrame df: A dataframe of SOLUTION/MATRIX_ESTIMATE.
    """

    # Matrix dimensions
    Q = m
    triangle = tri
    n = np.shape(Q)[0]

    if triangle == "L":

        # Make upper triangle NaNs
        Q[np.triu_indices(n, 1)] = np.nan

        row = []
        col = []
        q1 = []
        q2 = []
        q3 = []
        i = 0
        j = 0
        while i < n:
            while j <= i:
                row.append(i+1)
                col.append(j+1)
                q1.append(Q[i, j])
                q2.append(Q[i, j+1])
                q3.append(Q[i, j+2])
                j += 3
            j = 0
            i += 1

    if triangle == "U":

        # Make lower triangle NaNs
        Q[np.tril_indices(n, -1)] = np.nan

        row = []
        col = []
        q1 = []
        q2 = []
        q3 = []
        i = 0
        j = 0
        while i < n:
            while j < n:
                if j < n-2:
                    row.append(i+1)
                    col.append(j+1)
                    q1.append(Q[i, j])
                    q2.append(Q[i, j+1])
                    q3.append(Q[i, j+2])
                    j += 3

                if j == n-2: 
                    row.append(i+1)
                    col.append(j+1)
                    q1.append(Q[i, j])
                    q2.append(Q[i, j+1])
                    q3.append(np.nan)
                    j += 3

                if j == n-1: 
                    row.append(i+1)
                    col.append(j+1)
                    q1.append(Q[i, j])
                    q2.append(np.nan)
                    q3.append(np.nan)
                    j += 3
            
            i += 1
            j = i

    dict_temp = {
                "row":row,
                "col":col,
                "q1":q1,
                "q2":q2,
                "q3":q3,
    }
    df = pd.DataFrame(dict_temp)

    # Return
    return df

def dataframe2sinex_solution_matrix_estimate(df, tri="L"):
    """This function reads in a dataframe of the 
    SOLUTIONS/MATRIX_ESTIMATE block from a SINEX, the
    converts each row to a string in a list 
    ready for writing to SINEX.

    :param dataframe df: dataframe of SOLUTION/ESTIMATE block
    return: list of strings
    """
    
    # Upper or lower triangle
    triangle = tri

    lines = [f"+SOLUTION/MATRIX_ESTIMATE {triangle} COVA"]
    lines.append("*PARA1 PARA2 ____PARA2+0__________ ____PARA2+1__________ ____PARA2+2__________")

    for i in range(len(df.row)):
        p1 = df.row.iloc[i]
        p2 = df.col.iloc[i]
        q1 = df.q1.iloc[i]
        q2 = df.q2.iloc[i]
        q3 = df.q3.iloc[i]
        l = ' {:5n} {:5n} {:>21.14e} {:>21.14e} {:>21.14e}'.format(p1, p2, q1, q2, q3)
        l = l.replace('nan', '')
        lines.append(l)
    
    lines.append(f"-SOLUTION/MATRIX_ESTIMATE {triangle} COVA")

    # Return
    return lines

def writeSINEX(fp, 
               header=None, 
               comment=None, 
               siteID=None, 
               solutionEpochs=None, 
               solutionEstimate=None, 
               solutionMatrixEstimate=None,
               fileReference=None,
               inputAcknowledgments=None,
               solutionStatistics=None, 
               siteReceiver=None,
               siteAntenna=None,
               siteGpsPhaseCenter=None,
               siteEccentricity=None,
               solutionApriori=None,
               solutionMatrixApriori=None):
    """This function writes out SINEX blocks to a new SINEX file. The
    SINEX blocks can be obtained from many of the read_sinex_...() 
    functions when writing the same input to output. Or can use the 
    dataframe2sinex_...() functions when SINEX manipulations have
    occurred on that specific block. Input arguments are set to 'None'
    by default. This allows user to only write out the required blocks.

    Parameters:
    fp (string): Full filepath to output SINEX. 
    header (string): Single header line. Can get from read_sinex_header_line().
    comment (list of strings): +FILE/COMMENT block. Can get from read_sinex_comments().
    SiteID (list of strings): +SITE/ID block. Can get from read_sinex_site_id_block().
    SolutionEpochs (list of strings): +SOLUTION/EPOCHS block. Can get from read_sinex_solution_epochs_block().
    SolutionEstimatee (list of strings): +SOLUTION/ESTIMATE block. Can get from read_sinex_solution_estimate_block().
    SolutionMatricEstimate (list of strings): +SOLUTION/MATRIX_ESTIMATE block. Can get from read_sinex_solution_matrix_estimate_block().
    fileReference (list of strings): +FILE/REFERENCE block. Can get from read_sinex_file_reference_block().
    inputAcknowledgments (list of strings): +INPUT/ACKNOWLEDGEMENTS block. Can get from read_sinex_input_acknowledgments_block().
    siteReceiver (list of strings): +SITE/RECEIVER block. Can get from read_sinex_site_receiver_block().
    siteAntenna (list of strings): +SITE/ANTENNA block. Can get from read_sinex_site_antenna_block().
    siteGpsPhaseCenter (list of strings): +SITE/GPS_PHASE_CENTER block. Can get from read_site_gps_phase_center_block().
    siteEccentricity (list of strings): +SITE/ECCENTRICITY block. Can get from read_sinex_site_eccentricity_block().
    solutionApriori (list of strings): +SOLUTION/APRIORI block. Can get from read_sinex_solution_apriori_block().
    solutionMatrixApriori (list of strings): +SOLUTION/MATRIX_APRIORI block. Can get from read_sinex_solution_matric_apriori_block().

    Return:
    No return. But a new SINEX file will be written out to the file path (fp).

    """

    # Open File
    with open(fp, 'w') as f:

        # Header
        if header == None:
            pass
        else:
            f.write("{}".format(header))

            f.write("*-------------------------------------------------------------------------------\n")

        # Comment
        if comment == None:
            pass
        else:
            for i in range(len(comment)):
                f.write("{}\n".format(comment[i]))

            f.write("*-------------------------------------------------------------------------------\n")
        
        # FILE/REFERENCE
        if fileReference == None:
            pass
        else:
            for i in range(len(fileReference)):
                f.write("{}\n".format(fileReference[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # INPUT/ACKNOWLEDGMENTS
        if inputAcknowledgments == None:
            pass
        else:
            for i in range(len(inputAcknowledgments)):
                f.write("{}\n".format(inputAcknowledgments[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SOLUTION/STATISTICS
        if solutionStatistics == None:
            pass
        else:
            for i in range(len(solutionStatistics)):
                f.write("{}\n".format(solutionStatistics[i]))

            f.write("*-------------------------------------------------------------------------------\n")
        
        #SITE/ID
        if siteID == None:
            pass
        else:
            for i in range(len(siteID)):
                f.write("{}\n".format(siteID[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SITE/RECEIVER
        if siteReceiver == None:
            pass
        else:
            for i in range(len(siteReceiver)):
                f.write("{}\n".format(siteReceiver[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SITE/ANTENNA
        if siteAntenna == None:
            pass
        else:
            for i in range(len(siteAntenna)):
                f.write("{}\n".format(siteAntenna[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SITE/GPS_PHASE_CENTER
        if siteGpsPhaseCenter == None:
            pass
        else:
            for i in range(len(siteGpsPhaseCenter)):
                f.write("{}\n".format(siteGpsPhaseCenter[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SITE/ECCENTRICITY
        if siteEccentricity == None:
            pass
        else:
            for i in range(len(siteEccentricity)):
                f.write("{}\n".format(siteEccentricity[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SOLUTION/EPOCHS
        if solutionEpochs == None:
            pass
        else:
            for i in range(len(solutionEpochs)):
                f.write("{}\n".format(solutionEpochs[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SOLUTION/ESTIMATE
        if solutionEstimate == None:
            pass
        else:
            for i in range(len(solutionEstimate)):
                f.write("{}\n".format(solutionEstimate[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SOLUTION/APRIORI
        if solutionApriori == None:
            pass
        else:
            for i in range(len(solutionApriori)):
                f.write("{}\n".format(solutionApriori[i]))

            f.write("*-------------------------------------------------------------------------------\n")

        # SOLUTION/MATRIX_ESTIMATE
        if solutionMatrixEstimate == None:
            pass
        else:
            for i in range(len(solutionMatrixEstimate)):
                f.write("{}\n".format(solutionMatrixEstimate[i]))

            f.write("*-------------------------------------------------------------------------------\n")
        
        # SOLUTION/MATRIX_APRIORI
        if solutionMatrixApriori == None:
            pass
        else:
            for i in range(len(solutionMatrixApriori)):
                f.write("{}\n".format(solutionMatrixApriori[i]))
            
            f.write("*-------------------------------------------------------------------------------\n")

        # End Line
        f.write("%ENDSNX")

    f.close()

def remove_stns_sinex(sinex, sites):
    """This function removes a list sites from a SINEX file

    :param sinex: input SINEX file
    :param sites: list of the sites to be removed
    :return: SINEX file output.snx
    """

    # Open the output file
    with open('output.snx', 'w') as out:

        # Get header line and update the creation time and the number of
        # parameter estimates. Write the updated header line to the new file
        header = read_sinex_header_line(sinex)
        old_creation_time = header[15:27]
        creation_time = set_creation_time()
        header = header.replace(old_creation_time, creation_time)
        old_num_params = header[60:65]
        if header[70:71] == 'V':
            num_stn_params = 6
        else:
            num_stn_params = 3
        solution_epochs = read_sinex_solution_epochs_block(sinex)
        num_stns_to_remove = 0
        for line in solution_epochs:
            site = line[1:5]
            if site in sites:
                num_stns_to_remove += 1
        del solution_epochs
        num_params = int(old_num_params) - num_stn_params * num_stns_to_remove
        num_params = '{:05d}'.format(num_params)
        header = header.replace(str(old_num_params), str(num_params))
        out.write(header)

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +FILE/COMMENT block and write to output file
        comments = read_sinex_comments(sinex)
        for i in comments:
            out.write(f"{i}\n")

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the site ID block and write out the sites not being removed
        site_id = read_sinex_site_id_block(sinex)
        for line in site_id:
            if line.startswith('*') or line.startswith('+') or \
                    line.startswith('-'):
                out.write(f"{line}\n")
            else:
                site = line[1:5]
                if site not in sites:
                    out.write(f"{line}\n")
        del site_id

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the solution epochs block and write out the epochs of the
        # sites not being removed
        solution_epochs = read_sinex_solution_epochs_block(sinex)
        for line in solution_epochs:
            if line.startswith('*') or line.startswith('+') or \
                    line.startswith('-'):
                out.write(f"{line}\n")
            else:
                site = line[1:5]
                if site not in sites:
                    out.write(f"{line}\n")
        del solution_epochs

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the solution estimate block and write out the estimates of
        # the sites not being removed
        skip = []
        estimate_number = 0
        solution_estimate = read_sinex_solution_estimate_block(sinex)
        for line in solution_estimate:
            if line.startswith('*') or line.startswith('+') or \
                    line.startswith('-'):
                out.write(f"{line}\n")
            else:
                site = line[14:18]
                if site in sites:
                    num = int(line[0:6])
                    skip.append(num)
                else:
                    estimate_number += 1
                    number = '{:5d}'.format(estimate_number)
                    line = ' ' + number + line[6:]
                    out.write(f"{line}\n")
        del solution_estimate

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the matrix estimate block and write out minus the sites
        # being removed
        vcv = {}
        solution_matrix_estimate = \
            read_sinex_solution_matrix_estimate_block(sinex)
        if solution_matrix_estimate[0][26:27] == 'L':
            matrix = 'lower'
        elif solution_matrix_estimate[0][26:27] == 'U':
            matrix = 'upper'
        out.write(f"{solution_matrix_estimate[0]}\n")
        if solution_matrix_estimate[1].startswith('*'):
            out.write(f"{solution_matrix_estimate[1]}\n")
        for line in solution_matrix_estimate:
            if line.startswith(' '):
                cols = line.split()
                row = cols[0]
                for i in range(2, len(cols)):
                    try:
                        vcv[row].append(cols[i])
                    except KeyError:
                        vcv[row] = []
                        vcv[row].append(cols[i])
        block_end = solution_matrix_estimate[-1]
        del solution_matrix_estimate
        sub_vcv = {}
        sub_row = 0
        for i in range(1, len(vcv)+1):
            if i not in skip:
                sub_row += 1
                if matrix == 'lower':
                    for j in range(i):
                        if j+1 not in skip:
                            try:
                                sub_vcv[str(sub_row)].append(vcv[str(i)][j])
                            except KeyError:
                                sub_vcv[str(sub_row)] = []
                                sub_vcv[str(sub_row)].append(vcv[str(i)][j])
                if matrix == 'upper':
                    for j in range(len(vcv)-(i-1)):
                        if j+i not in skip:
                            try:
                                sub_vcv[str(sub_row)].append(vcv[str(i)][j])
                            except KeyError:
                                sub_vcv[str(sub_row)] = []
                                sub_vcv[str(sub_row)].append(vcv[str(i)][j])
        for i in range(1, len(sub_vcv)+1):
            para1 = '{:5d}'.format(i)
            if matrix == 'lower':
                j = -2
            elif matrix == 'upper':
                j = i - 3
            while sub_vcv[str(i)]:
                j += 3
                para2 = '{:5d}'.format(j)
                line = ' ' + para1 + ' ' + para2
                n = min([3, len(sub_vcv[str(i)])])
                for k in range(n):
                    val = '{:21.14e}'.format(float(sub_vcv[str(i)].pop(0)))
                    line += ' ' + str(val)
                out.write(line + '\n')
        out.write(block_end)

        # Write out the trailer line
        out.write('%ENDSNX\n')

    return

def remove_velocity_sinex(sinex):
    """This function reads in a SINEX file and removes the 
    velocity parameters, including the zeros of the SOLUTION/MATRIX_ESTIMATE block,

    :param str sinex: input SINEX file
    return: SINEX file output.snx
    """

    # From header, confirm that the SINEX has velocity parameters
    header = read_sinex_header_line(sinex)
    header = header.strip()
    if header[-1] == 'V':
        pass
    else:
        print("Not removing velocities because SINEX file does not have velocity parameters.")
        exit()

    # Open the output file
    with open('output.snx', 'w') as out:

        # With header line: 
        # - update the creation time
        # - update number of parameter estimates
        # - remove 'V' from parameter list 
        # - then write to file
        old_creation_time = header[15:27]
        creation_time = set_creation_time()
        header = header.replace(old_creation_time, creation_time)
        old_num_params = int(header[60:65])
        num_params = int(old_num_params / 2)
        header = header.replace(str(old_num_params), str(num_params))
        header = header.replace('V', '')
        out.write(header)
        out.write("\n")
        del header

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +FILE/COMMENT block and write to output file
        comments = read_sinex_comments(sinex)
        for i in comments:
            out.write(f"{i}\n")

        out.write("*-------------------------------------------------------------------------------\n")
        
        # Read in the +SITE/ID block and write to file
        site_id = read_sinex_site_id_block(sinex)
        for line in site_id:
            out.write(f"{line}\n")  
        del site_id

        out.write("*-------------------------------------------------------------------------------\n")
        
        # Read in the +SOLUTION/EPOCHS block and write to file
        # - also collect count on number of sites for use later
        numSites = 0
        solution_epochs = read_sinex_solution_epochs_block(sinex)
        for line in solution_epochs:
            out.write(f"{line}\n")
            if line[0]!="+" and line[0]!="*" and line[0]!="-":
                numSites+=1
        del solution_epochs

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +SOLUTION/ESTIMATE block: 
        # - gather velocity indices 
        # - remove velocity rows 
        # - change indices to account for removed velocities
        # - write to file
        vel_indices = []
        estimate_number = 0
        solution_estimate = read_sinex_solution_estimate_block(sinex)
        for line in solution_estimate:
            if line[7:10]=="VEL":
                vel_indices.append(int(line[0:6]))
            elif line[0]=="+" or line[0]=="*" or line[0]=="-":
                out.write(f"{line}\n")
            else:
                estimate_number+=1
                number = '{:5d}'.format(estimate_number)
                line = ' ' + number + line[6:]
                out.write(f"{line}\n")
        del solution_estimate

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +SOLUTION/MATRIX_ESTIMATE block:
        # - identify if matrix is upper or lower triangle form
        # - form full matrix
        # - remove all rows/columns associated with velocity
        # - write to file with new matrix indices (para1, para2)
        solution_matrix_estimate = read_sinex_solution_matrix_estimate_block(sinex)
        if solution_matrix_estimate[0][26:27] == 'L':
                matrix = 'lower'
        elif solution_matrix_estimate[0][26:27] == 'U':
                matrix = 'upper'
        # Size of original matrix
        matrix_dim = numSites*6
        # Setup matrix of zeros
        Q = zeros((matrix_dim, matrix_dim))
        # Write initial comment line(s), save last comment line, and form matrix
        for line in solution_matrix_estimate:
            if line[0]=="+":
                out.write(f"{line}\n")
                continue
            if line[0]=="*":
                out.write(f"{line}\n")
                continue
            if line[0]=="-":
                block_end = line
            else:
                lineMAT = re.split('\s+', line)
                lineMAT = list(filter(None, lineMAT))
                row = int(lineMAT[0])
                col = int(lineMAT[1])
            if len(lineMAT) >= 3:
                q1 = float(lineMAT[2])
                Q[row-1, col-1] = q1
                Q[col-1, row-1] = q1
            if len(lineMAT) >= 4:
                q2 = float(lineMAT[3])
                Q[row-1, col] = q2
                Q[col, row-1] = q2
            if len(lineMAT) >= 5:
                q3 = float(lineMAT[4])
                Q[row-1, col+1] = q3
                Q[col+1, row-1] = q3
        # Remove velocity rows/columns from matrix
        num_removed = 0
        for i in vel_indices:
            Q = delete(Q, i-1-num_removed, 0)
            Q = delete(Q, i-1-num_removed, 1)
            num_removed += 1
        # Write matrix to SINEX (lower triangle)
        if matrix == 'lower':
            i = 0
            j = 0
            while i < len(Q):
                while j <= i:
                    out.write(f" {i+1:5d} {j+1:5d} {Q[i,j]:21.14E} ")
                    j += 1
                    if j <= i:
                        out.write(f"{Q[i,j]:21.14E} ")
                        j += 1
                    if j <= i:
                        out.write(f"{Q[i,j]:21.14E}")
                        j += 1
                    out.write(" \n")
                j = 0
                i += 1
        # Write matrix to SINEX (upper triangle)
        if matrix == 'upper':
            j = 0
            for i in range(len(Q)):
                j = i 
                while j < len(Q):
                    out.write(f" {i+1:5d} {j+1:5d} {Q[i,j]:21.14E} ")
                    j += 1
                    if j < len(Q):
                        out.write(f"{Q[i,j]:21.14E} ")
                        j += 1
                    if j < len(Q):
                        out.write(f"{Q[i,j]:21.14E}")
                        j += 1
                    out.write(" \n")
        # Write out end of block line, and delete large variables
        out.write(block_end)
        del solution_matrix_estimate
        del Q

        # Write out the trailer line
        out.write('%ENDSNX\n')

    return

def remove_matrixzeros_sinex(sinex):
    """This function reads in a SINEX file and removes the 
    zeros from the SOLUTION/MATRIX_ESTIMATE block only. 

    :param str sinex: input SINEX file
    return: SINEX file output.snx
    """
    
    # Open the output file
    with open('output.snx', 'w') as out:

        # With header line:
        # - update the creation time 
        # - then write to file
        header = read_sinex_header_line(sinex)
        old_creation_time = header[15:27]
        creation_time = set_creation_time()
        header = header.replace(old_creation_time, creation_time)
        out.write(header)
        del header

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +FILE/COMMENT block and write to output file
        comments = read_sinex_comments(sinex)
        for i in comments:
            out.write(f"{i}\n")

        out.write("*-------------------------------------------------------------------------------\n")
        
        # Read in the +SITE/ID block and write to file
        site_id = read_sinex_site_id_block(sinex)
        for line in site_id:
            out.write(f"{line}\n")  
        del site_id

        out.write("*-------------------------------------------------------------------------------\n")
        
        # Read in the +SOLUTION/EPOCHS block and write to file
        solution_epochs = read_sinex_solution_epochs_block(sinex)
        for line in solution_epochs:
            out.write(f"{line}\n")
        del solution_epochs

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +SOLUTION/ESTIMATE block
        solution_estimate = read_sinex_solution_estimate_block(sinex)
        for line in solution_estimate:
                out.write(f"{line}\n")
        del solution_estimate

        out.write("*-------------------------------------------------------------------------------\n")

        # Read in the +SOLUTION/MATRIX_ESTIMATE block:
        # - Remove lines that contain only zeros
        solution_matrix_estimate = read_sinex_solution_matrix_estimate_block(sinex)
        for line in solution_matrix_estimate:
            col = line.split()
            numCol = len(col)
            if numCol==3:
                if col[2]=="0.00000000000000e+00":
                    continue
            if numCol==4:
                if col[2]=="0.00000000000000e+00" and col[3]=="0.00000000000000e+00":
                    continue
            if numCol==5:
                if col[2]=="0.00000000000000e+00" and col[3]=="0.00000000000000e+00" and col[4]=="0.00000000000000e+00":
                    continue
            out.write(line)
        del solution_matrix_estimate

        # Write out the trailer line
        out.write('%ENDSNX\n')

    return



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Convert Module
"""

from math import (sin, cos, atan2, radians, degrees,
                  sqrt, cosh, sinh, tan, atan, log)
import datetime
import warnings
from geodepy.constants import utm, isg, grs80, ans
from geodepy.angles import (DECAngle, HPAngle, GONAngle, DMSAngle, DDMAngle,
                            dec2hp, dec2hpa, dec2gon, dec2gona,
                            dec2dms, dec2ddm,
                            hp2dec, hp2deca, hp2gon, hp2gona,
                            hp2dms, hp2ddm, hp2rad, hp2dec_v,
                            gon2dec, gon2deca, gon2hp, gon2hpa,
                            gon2dms, gon2ddm, gon2rad,
                            dd2sec, angular_typecheck)


def polar2rect(r, theta):
    """
    Converts point in polar coordinates to corresponding rectangular coordinates
    Theta is degrees and is measured clockwise from the positive y axis
    (i.e. north)
    :param r: Radius
    :param theta: Angle (decimal degrees)
    :type theta: Float (decimal degrees), DMSAngle or DDMAngle
    :return: Rectangular Coordinates X, Y
    """
    theta = angular_typecheck(theta)
    x = r * sin(radians(theta))
    y = r * cos(radians(theta))
    return x, y


def rect2polar(x, y):
    """
    Converts point in rectangular coordinates to corresponding polar coordinates
    Angular component of polar coordinates (theta) is in decimal degrees and
    is measured clockwise from the positive y axis (i.e. north)
    :param x: Rectangular Coordinate X
    :param y: Rectangular Coordinate Y
    :return r: Radius
    :return theta: Angle (decimal degrees)
    :rtype theta: float (decimal degrees)
    """
    r = sqrt(x ** 2 + y ** 2)
    theta = atan2(x, y)
    if theta < 0:
        theta = degrees(theta) + 360
    else:
        theta = degrees(theta)
    return r, theta


def rect_radius(ellipsoid):
    """
    Computes the Rectifying Radius of an Ellipsoid with specified Inverse
    Flattening (See Ref 2 Equation 3)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Ellipsoid Rectifying Radius
    """
    nval = (1 / float(ellipsoid.inversef)) /\
           (2 - (1 / float(ellipsoid.inversef)))
    nval2 = nval**2
    return (ellipsoid.semimaj / (1 + nval) * ((nval2 *
                                              (nval2 *
                                               (nval2 *
                                                (25 * nval2 + 64)
                                                + 256)
                                               + 4096)
                                              + 16384)
                                              / 16384.))


def alpha_coeff(ellipsoid):
    """
    Computes the set of Alpha coefficients of an Ellipsoid with specified
    Inverse Flattening (See Ref 2 Equation 5)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Alpha coefficients a2, a4 ... a16
    :rtype: tuple
    """
    nval = ellipsoid.n
    a2 = ((nval *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                ((37884525 - 75900428 * nval)
                 * nval + 42422016)
                - 89611200)
               + 46287360)
              + 63504000)
             - 135475200)
            + 101606400))
          / 203212800.)

    a4 = ((nval ** 2 *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                (148003883 * nval + 83274912)
                - 178508970)
               + 77690880)
              + 67374720)
             - 104509440)
            + 47174400))
          / 174182400.)

    a6 = ((nval ** 3 *
           (nval *
            (nval *
             (nval *
              (nval *
               (318729724 * nval - 738126169)
               + 294981280)
              + 178924680)
             - 234938880)
            + 81164160))
          / 319334400.)

    a8 = ((nval ** 4 *
           (nval *
            (nval *
             ((14967552000 - 40176129013 * nval) * nval + 6971354016)
             - 8165836800)
            + 2355138720))
          / 7664025600.)

    a10 = ((nval ** 5 *
            (nval *
             (nval *
              (10421654396 * nval + 3997835751)
              - 4266773472)
             + 1072709352))
           / 2490808320.)

    a12 = ((nval ** 6 *
            (nval *
             (175214326799 * nval - 171950693600)
             + 38652967262))
           / 58118860800.)

    a14 = ((nval ** 7 *
            (13700311101 - 67039739596 * nval))
           / 12454041600.)

    a16 = (1424729850961 * nval ** 8) / 743921418240.
    return a2, a4, a6, a8, a10, a12, a14, a16


def beta_coeff(ellipsoid):
    """
    Computes the set of Beta coefficients of an Ellipsoid with specified
    Inverse Flattening (See Ref 2 Equation 23)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Alpha coefficients a2, a4 ... a16
    :rtype: tuple
    """
    nval = ellipsoid.n
    b2 = ((nval *
           (nval *
            (nval *
             (nval *
              (nval *
               (nval *
                ((37845269 - 31777436 * nval) - 43097152)
                + 42865200)
               + 752640)
              - 104428800)
             + 180633600)
            - 135475200))
          / 270950400.)

    b4 = ((nval ** 2 *
           (nval *
            (nval *
             (nval *
              (nval *
               ((-24749483 * nval - 14930208) * nval + 100683990)
               - 152616960)
              + 105719040)
             - 23224320)
            - 7257600))
          / 348364800.)

    b6 = ((nval ** 3 *
           (nval *
            (nval *
             (nval *
              (nval *
               (232468668 * nval - 101880889)
               - 39205760)
              + 29795040)
             + 28131840)
            - 22619520))
          / 638668800.)

    b8 = ((nval ** 4 *
           (nval *
            (nval *
             ((-324154477 * nval - 1433121792) * nval + 876745056)
             + 167270400)
            - 208945440))
          / 7664025600.)

    b10 = ((nval ** 5 *
            (nval *
             (nval *
              (312227409 - 457888660 * nval)
              + 67920528)
             - 70779852))
           / 2490808320.)

    b12 = ((nval ** 6 *
            (nval *
             (19841813847 * nval + 3665348512)
             - 3758062126))
           / 116237721600.)

    b14 = ((nval ** 7 *
            (1989295244 * nval - 1979471673))
           / 49816166400.)

    b16 = ((-191773887257 * nval ** 8) / 3719607091200.)
    return b2, b4, b6, b8, b10, b12, b14, b16


def psfandgridconv(xi1, eta1, lat, lon, cm, conf_lat, ellipsoid=grs80, prj=utm):
    """
    Calculates Point Scale Factor and Grid Convergence. Used in convert.geo2grid
    and convert.grid2geo
    :param xi1: Transverse Mercator Ratio Xi
    :param eta1: Transverse Mercator Ratio Eta
    :param lat: Latitude
    :type lat: Decimal Degrees, DMSAngle or DDMAngle
    :param lon: Longitude
    :type lon: Decimal Degrees, DMSAngle or DDMAngle
    :param cm: Central Meridian
    :param conf_lat: Conformal Latitude
    :param ellipsoid: Ellipsoid Object (default: GRS80)
    :return: Point Scale Factor, Grid Convergence (Decimal Degrees)
    :rtype: tuple
    """
    lat = angular_typecheck(lat)
    lon = angular_typecheck(lon)
    A = rect_radius(ellipsoid)
    a = alpha_coeff(ellipsoid)
    lat = radians(lat)
    long_diff = radians(lon - cm)

    # Point Scale Factor
    p = 1
    q = 0
    for r in range(1, 9):
        p += 2*r * a[r-1] * cos(2*r * xi1) * cosh(2*r * eta1)
        q += 2*r * a[r-1] * sin(2*r * xi1) * sinh(2*r * eta1)
    q = -q
    psf = (float(prj.cmscale)
           * (A / ellipsoid.semimaj)
           * sqrt(q**2 + p**2)
           * ((sqrt(1 + (tan(lat)**2))
               * sqrt(1 - ellipsoid.ecc1sq * (sin(lat)**2)))
              / sqrt((tan(conf_lat)**2) + (cos(long_diff)**2))))

    # Grid Convergence
    grid_conv = degrees(atan(abs(q / p))
                        + atan(abs(tan(conf_lat) * tan(long_diff))
                               / sqrt(1 + tan(conf_lat)**2)))
    if cm > lon and lat < 0:
        grid_conv = -grid_conv
    elif cm < lon and lat > 0:
        grid_conv = -grid_conv

    return psf, grid_conv


def geo2grid(lat, lon, zone=0, ellipsoid=grs80, prj=utm):
    """
    Takes a geographic co-ordinate (latitude, longitude) and returns its
    corresponding Hemisphere, Zone and Projection Easting and Northing, Point
    Scale Factor and Grid Convergence. Default Projection is Universal
    Transverse Mercator Projection using
    GRS80 Ellipsoid parameters.
    :param lat: Latitude
    :type lat: Float (Decimal Degrees), DMSAngle or DDMAngle
    :param lon: Longitude
    :type lon: Float (Decimal Degrees, DMSAngle or DDMAngle
    :param zone: Optional Zone Number - Only required if calculating grid
                                        co-ordinate outside zone boundaries
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: hemisphere, zone, east (m), north (m), Point Scale Factor,
             Grid Convergence (Decimal Degrees)
    :rtype: tuple
    """

    # Convert DMSAngle and DDMAngle to Decimal Angle
    lat = angular_typecheck(lat)
    lon = angular_typecheck(lon)
    # Input Validation - UTM Extents and Values
    zone = int(zone)
    if prj == isg:
        if zone not in (0, 541, 542, 543, 551, 552, 553, 561, 562, 563, 572):
            raise ValueError('Invalid Zone - Choose from 541, 542, 543, 551, 552, 553, 561, 562, 563, 572')
    else:
        if zone < 0 or zone > 60:
                raise ValueError('Invalid Zone - Zones from 1 to 60')

    if lat < -80 or lat > 84:
        raise ValueError('Invalid Latitude - Latitudes from -80 to +84')

    if lon < -180 or lon > 180:
        raise ValueError('Invalid Longitude - Longitudes from -180 to +180')

    if prj == isg and ellipsoid != ans:
        warnings.warn(message='ISG projection should be used with ANS ellipsoid', category=UserWarning)

    A = rect_radius(ellipsoid)
    a = alpha_coeff(ellipsoid)
    lat = radians(lat)
    # Calculate Zone
    if zone == 0:
        if prj == isg:
            amgzone = (float(lon) - (prj.initialcm - (4.5 * prj.zonewidth))) / (prj.zonewidth * 3)
            subzone = int((amgzone - int(amgzone)) * 3 + 1)
            amgzone = int(amgzone)
            zone = int(f'{amgzone}{subzone}')
        else:
            zone = int((float(lon) - (prj.initialcm - (1.5 * prj.zonewidth))) / prj.zonewidth)
    if prj == isg:
        amgzone = int(str(zone)[:2])
        subzone = int(str(zone)[2])
        cm = float((amgzone - 1) * prj.zonewidth * 3 + prj.initialcm + (subzone - 2) * prj.zonewidth)
    else:
        cm = float(zone * prj.zonewidth + prj.initialcm - prj.zonewidth)

    # Conformal Latitude
    sigx = (ellipsoid.ecc1 * tan(lat)) / sqrt(1 + (tan(lat) ** 2))
    sig = sinh(ellipsoid.ecc1 * (0.5 * log((1 + sigx) / (1 - sigx))))
    conf_lat = tan(lat) * sqrt(1 + sig ** 2) - sig * sqrt(1 + (tan(lat) ** 2))
    conf_lat = atan(conf_lat)

    # Longitude Difference
    long_diff = radians(lon - cm)
    # Gauss-Schreiber Ratios
    xi1 = atan(tan(conf_lat) / cos(long_diff))
    eta1x = sin(long_diff) / (sqrt(tan(conf_lat) ** 2 + cos(long_diff) ** 2))
    eta1 = log(eta1x + sqrt(1 + eta1x ** 2))

    # Transverse Mercator Ratios
    eta = eta1
    xi = xi1
    for r in range(1, 9):
        eta += a[r-1] * cos(2*r * xi1) * sinh(2*r * eta1)
        xi += a[r-1] * sin(2*r * xi1) * cosh(2*r * eta1)

    # Transverse Mercator Co-ordinates
    x = A * eta
    y = A * xi

    # Hemisphere-dependent UTM Projection Co-ordinates
    east = prj.cmscale * x + prj.falseeast
    if y < 0:
        hemisphere = 'South'
        north = prj.cmscale * y + prj.falsenorth
    else:
        hemisphere = 'North'
        falsenorth = 0
        north = prj.cmscale * y + falsenorth

    # Point Scale Factor and Grid Convergence
    psf, grid_conv = psfandgridconv(xi1, eta1, degrees(lat), lon, cm, conf_lat)

    return (hemisphere, zone,
            round(float(east), 4),
            round(float(north), 4),
            round(psf, 8), grid_conv)


def grid2geo(zone, east, north, hemisphere='south', ellipsoid=grs80, prj=utm):
    """
    Takes a Transverse Mercator grid co-ordinate (Zone, Easting, Northing,
    Hemisphere) and returns its corresponding Geographic Latitude and Longitude,
    Point Scale Factor and Grid Convergence. Default Projection is Universal
    Transverse Mercator Projection using GRS80 Ellipsoid parameters.
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param hemisphere: String - 'North' or 'South'(default)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Latitude and Longitude (Decimal Degrees), Point Scale Factor,
             Grid Convergence (Decimal Degrees)
    :rtype: tuple
    """
    # Input Validation - UTM Extents and Values
    zone = int(zone)
    if prj == isg:
        if zone not in (541, 542, 543, 551, 552, 553, 561, 562, 563, 572):
            raise ValueError('Invalid Zone - Choose from 541, 542, 543, 551, 552, 553, 561, 562, 563, 572')
    else:
        if zone < 0 or zone > 60:
            raise ValueError('Invalid Zone - Zones from 1 to 60')

    if east < -2830000 or east > 3830000:
        raise ValueError('Invalid Easting - Must be within'
                         '3330km of Central Meridian')

    if north < 0 or north > 10000000:
        raise ValueError('Invalid Northing - Must be between 0 and 10,000,000m')

    h = hemisphere.lower()
    if h != 'north' and h != 'south':
        raise ValueError('Invalid Hemisphere - String, either North or South')

    if prj == isg and ellipsoid != ans:
        warnings.warn(message='ISG projection should be used with ANS ellipsoid', category=UserWarning)

    A = rect_radius(ellipsoid)
    b = beta_coeff(ellipsoid)
    # Transverse Mercator Co-ordinates
    x = (east - float(prj.falseeast)) / float(prj.cmscale)
    if hemisphere.lower() == 'north':
        y = -(north / float(prj.cmscale))
        hemisign = -1
    else:
        y = (north - float(prj.falsenorth)) / float(prj.cmscale)
        hemisign = 1

    # Transverse Mercator Ratios
    xi = y / A
    eta = x / A

    # Gauss-Schreiber Ratios
    eta1 = eta
    xi1 = xi
    for r in range(1, 9):
        eta1 += b[r-1] * cos(2*r * xi) * sinh(2*r * eta)
        xi1 += b[r-1] * sin(2*r * xi) * cosh(2*r * eta)

    # Conformal Latitude
    conf_lat = (sin(xi1)) / (sqrt((sinh(eta1)) ** 2 + (cos(xi1)) ** 2))
    t1 = conf_lat
    conf_lat = atan(conf_lat)

    # Finding t using Newtons Method
    def sigma(tn, ecc1):
        return (sinh(ecc1
                     * 0.5
                     * log((1 + ((ecc1 * tn) / (sqrt(1 + tn ** 2))))
                           / (1 - ((ecc1 * tn) / (sqrt(1 + tn ** 2)))))))

    def ftn(tn, ecc1):
        return (t * sqrt(1 + (sigma(tn, ecc1)) ** 2) -
                sigma(tn, ecc1) * sqrt(1 + tn ** 2) - t1)

    def f1tn(tn, ecc1, ecc1sq):
        return ((sqrt(1 + (sigma(tn, ecc1)) ** 2) * sqrt(1 + tn ** 2)
                 - sigma(tn, ecc1) * tn)
                * (((1 - float(ecc1sq)) * sqrt(1 + t ** 2))
                   / (1 + (1 - float(ecc1sq)) * t ** 2)))

    diff = 1
    t = t1
    itercount = 0
    while diff > 1e-15 and itercount < 100:
        itercount += 1
        t_before = t
        t = t - (ftn(t, ellipsoid.ecc1)
                 / f1tn(t, ellipsoid.ecc1, ellipsoid.ecc1sq))
        diff = abs(t - t_before)
    lat = degrees(atan(t))

    # Compute Longitude
    if prj == isg:
        amgzone = int(str(zone)[:2])
        subzone = int(str(zone)[2])
        cm = float((amgzone - 1) * prj.zonewidth * 3 + prj.initialcm + (subzone - 2) * prj.zonewidth)
    else:
        cm = float((zone * prj.zonewidth) + prj.initialcm - prj.zonewidth)
    long_diff = degrees(atan(sinh(eta1) / cos(xi1)))
    long = cm + long_diff

    # Point Scale Factor and Grid Convergence
    psf, grid_conv = psfandgridconv(xi1, eta1, lat, long, cm, conf_lat)

    return (hemisign * round(lat, 11),
            round(long, 11), round(psf, 8),
            hemisign * grid_conv)


def xyz2llh(x, y, z, ellipsoid=grs80):
    """
    Converts Cartesian X, Y, Z coordinate to Geographic Latitude, Longitude and
    Ellipsoid Height. Default Ellipsoid parameters used are GRS80.
    :param x: Cartesian X Coordinate (metres)
    :param y: Cartesian Y Coordinate (metres)
    :param z: Cartesian Z Coordinate (metres)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Geographic Latitude (Decimal Degrees), Longitude (Decimal Degrees)
    and Ellipsoid Height (metres)
    :rtype: tuple
    """
    # Calculate Longitude
    long = atan2(y, x)
    # Calculate Latitude
    p = sqrt(x**2 + y**2)
    latinit = atan((z*(1+ellipsoid.ecc2sq))/p)
    lat = latinit
    itercheck = 1
    while abs(itercheck) > 1e-10:
        nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat))**2))
        itercheck = lat - atan((z + nu * ellipsoid.ecc1sq * sin(lat))/p)
        lat = atan((z + nu * ellipsoid.ecc1sq * sin(lat))/p)
    nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat))**2))
    ellht = p/(cos(lat)) - nu
    # Convert Latitude and Longitude to Degrees
    lat = degrees(lat)
    long = degrees(long)
    return lat, long, ellht


def llh2xyz(lat, lon, ellht=0, ellipsoid=grs80):
    """
    Converts Geographic Latitude, Longitude and Ellipsoid Height to Cartesian
    X, Y and Z Coordinates. Default Ellipsoid parameters used are GRS80.
    :param lat: Geographic Latitude
    :type lat: Float (Decimal Degrees), DMSAngle or DDMAngle
    :param lon: Geographic Longitude
    :type lon: Float (Decimal Degrees), DMSAngle or DDMAngle
    :param ellht: Ellipsoid Height (metres, default is 0m)
    :param ellipsoid: Ellipsoid Object
    :type ellipsoid: Ellipsoid
    :return: Cartesian X, Y, Z Coordinate in metres
    :rtype: tuple
    """
    # Convert lat & long to radians
    lat = radians(angular_typecheck(lat))
    lon = radians(angular_typecheck(lon))
    # Calculate Ellipsoid Radius of Curvature in the Prime Vertical - nu
    if lat == 0:
        nu = grs80.semimaj
    else:
        nu = ellipsoid.semimaj/(sqrt(1 - ellipsoid.ecc1sq * (sin(lat)**2)))
    # Calculate x, y, z
    x = (nu + ellht) * cos(lat) * cos(lon)
    y = (nu + ellht) * cos(lat) * sin(lon)
    z = ((ellipsoid.semimin**2 / ellipsoid.semimaj**2) * nu + ellht) * sin(lat)
    return x, y, z


def date_to_yyyydoy(date):
    """
    Convert a datetime.date object to a string in the form 'yyyy.doy',
    where yyyy is the 4 character year number and doy is the 3 character
    day of year
    :param date: datetime.date object
    :type date: datetime.date
    :return: string with date in the form 'yyyy.doy'
    :rtype: str
    """
    try:
        return (str(date.timetuple().tm_year) + '.' +
                str(date.timetuple().tm_yday).zfill(3))
    except AttributeError:
        raise AttributeError('Invalid date: date must be datetime.date object')


def yyyydoy_to_date(yyyydoy):
    """
    Convert a string in the form of either 'yyyydoy' or 'yyyy.doy' to a
    datetime.date object, where yyyy is the 4 character year number and doy
    is the 3 character day of year
    :param yyyydoy: string with date in the form 'yyyy.doy' or 'yyyydoy'
    :return: datetime.date object
    :rtype: datetime.date
    """
    try:
        if '.' in yyyydoy:
            if len(yyyydoy) != 8:
                raise ValueError('Invalid string: must be yyyydoy or yyyy.doy')
            yyyy, doy = yyyydoy.split('.')
        else:
            if len(yyyydoy) != 7:
                raise ValueError('Invalid string: must be yyyydoy or yyyy.doy')
            yyyy = yyyydoy[0:4]
            doy = yyyydoy[4:7]
        return datetime.date(int(yyyy), 1, 1) + datetime.timedelta(int(doy) - 1)
    except ValueError:
        raise ValueError('Invalid string: must be yyyydoy or yyyy.doy')
